# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, MutableMapping, Optional, Sequence, Tuple, Union

import jax
from jaxtyping import Array, Float

from gpfy.training import TrainState

PRNG = Union[jax.random.PRNGKeyArray, Array]
ActiveDims = Union[slice, Sequence[int]]
Collection = MutableMapping[str, Any]
VariableDict = MutableMapping[str, Collection]
BijectorDict = MutableMapping[str, Collection]
ConstantDict = MutableMapping[str, Collection]
TrainableDict = MutableMapping[str, Collection]

TrainingData = Tuple[Float[Array, "N Din"], Float[Array, "N Dout"]]
TrainingDataDict = Dict[str, Float[Array, "N D"]]
TrainStepFn = Callable[[TrainState, Optional[Array]], Tuple[TrainState, float]]



# Copyright 2023 Stefanos Eleftheriadis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterator
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
from datasets import Dataset
from jaxtyping import Array

from gpfy.likelihoods import Likelihood
from gpfy.model import GP
from gpfy.param import Param
from gpfy.training import TrainState
from gpfy.typing import TrainingData, TrainingDataDict, TrainStepFn
from gpfy.variational import VariationalDistribution


def batched_data_generator(dataset: Dataset, batch_size: int) -> Iterator[TrainingDataDict]:
    """
    Create a data-generator for minibatching using a HuggingFace Dataset.

    NOTE: Currently there is no way to control the seed for the random shuffling via passing a
    `PRNG` to `Dataset`.

    Args:
        dataset: The dataset to wrap in the generator.
        batch_size: The size of the minibatch.

    Yields:
        A `TrainingDataDict` with the arrays the size of `batch_size`.
    """

    def _reset() -> Dataset:
        """Reset the iterator once it is exhausted."""
        # currently shuffling with a PRNG in a jit function is not working.
        # key, subkey = jr.split(key)
        # return dataset.shuffle(subkey.tolist()).iter(batch_size=batch_size), key
        return dataset.shuffle().iter(batch_size=batch_size, drop_last_batch=True)

    gen = _reset()
    while True:
        try:
            yield next(gen)
        except StopIteration:
            gen = _reset()


def create_training_step(
    model: GP,
    dataset: Dataset,
    dataset_xy_keys: Tuple[str, str],
    q: VariationalDistribution,
    lik: Likelihood,
    batch_size=None,
) -> TrainStepFn:
    """
    Create a training step callable that we can use in a `jax.lax.scan` to optimise the model.

    Args:
        model: A GP model.
        dataset: A HuggingFace dataset to fit with the GP model.
        dataset_xy_keys: The keys to the dataset dictionary that correspond to the `X` and `Y` data.
        q: The variational distribution.
        lik: The likelihood.
        batch_size: The size of the minibatch. Default to None, which means we fit the entire data.

    Returns:
        A training step callable that expects a `TrainState` input to call during optimisation.
    """
    x_key, y_key = dataset_xy_keys
    if batch_size:
        N = dataset.num_rows
        data_gen = batched_data_generator(dataset, batch_size=batch_size)

        @jax.jit
        def train_step(
            state: TrainState, dummy_input_for_scan: Optional[Array] = None
        ) -> Tuple[TrainState, float]:
            data = next(data_gen)
            X, Y = data[x_key], data[y_key]

            def loss_fn(params):
                free_param = state.apply_fn(params)
                return -elbo(free_param.constrained(), model, q, lik, (X, Y), dataset_size=N)

            loss_val, grads = jax.value_and_grad(loss_fn)(state.params)
            new_state = state.apply_gradients(grads=grads)

            return new_state, loss_val

        return train_step
    else:

        @jax.jit
        def train_step(
            state: TrainState, dummy_input_for_scan: Optional[Array] = None
        ) -> Tuple[TrainState, float]:
            X, Y = dataset[x_key], dataset[y_key]

            def loss_fn(params):
                free_param = state.apply_fn(params)
                return -elbo(free_param.constrained(), model, q, lik, (X, Y))

            loss_val, grads = jax.value_and_grad(loss_fn)(state.params)
            new_state = state.apply_gradients(grads=grads)

            return new_state, loss_val

        return train_step


@jax.jit
def elbo(
    param: Param,
    m: GP,
    q: VariationalDistribution,
    lik: Likelihood,
    train_data: TrainingData,
    dataset_size: int = -1,
) -> float:
    """
    The variational lower bound for inference in sparse Gaussian process.

    ELBO = Σ𝔼_q[log𝒩(p(y|f)] - KL[q(u)||p(u)]

    Args:
        param: A `Param` initialised with the model.
        m: The GP model.
        q: The variational distribution.
        lik: The likelihood.
        train_data: A tuple containing the training data.
        dataset_size: The full dataset size, in case we do minibatchgin. Defaults to -1, for no
        minibatching inference.

    Returns:
        The evidence lower bound.
    """
    X, Y = train_data
    fmu, fvar = m.predict_diag(param, X)
    var_exp = lik.variational_expectations(param, fmu, fvar)(Y)
    scale = jax.lax.cond(dataset_size > 0, lambda: dataset_size, lambda: jnp.shape(X)[0])
    KL = q.prior_KL(param)
    return scale * jnp.mean(var_exp, -1) - KL



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Union

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from jax import custom_vjp
from jax.tree_util import Partial
from jaxtyping import Array, Bool, Float

from gpfy.utils import dataclass, field


@jax.jit
def _gegenbauer(level: int, alpha: Union[float, Float[Array, ""]], x: Array) -> Array:
    """
    Compute the gegenbauer polynomial Cᵅₙ(x) recursively.

    Cᵅ₀(x) = 1
    Cᵅ₁(x) = 2αx
    Cᵅₙ(x) = (2x(n + α - 1) Cᵅₙ₋₁(x) - (n + 2α - 2) Cᵅₙ₋₂(x)) / n

    Args:
        level: The order of the polynomial.
        alpha: The hyper-sphere constant given by (d - 2) / 2 for the Sᵈ⁻¹ sphere.
        x: Input array.

    Returns:
        The Gegenbauer polynomial evaluated at `x`.
    """
    C_0 = jnp.ones_like(x, dtype=x.dtype)
    C_1 = 2 * alpha * x

    def step(Cs_and_n: Tuple[Array, Array, Float[Array, ""]]) -> Tuple[Array, Array, Array]:
        C, C_prev, n = Cs_and_n
        C, C_prev = (2 * x * (n + alpha - 1) * C - (n + 2 * alpha - 2) * C_prev) / n, C
        return C, C_prev, n + 1

    def cond(Cs_and_n: Tuple[Array, Array, Float[Array, ""]]) -> Bool[Array, ""]:
        n = Cs_and_n[2]
        return n <= level

    return jax.lax.cond(
        level == 0,
        lambda: C_0,
        lambda: jax.lax.while_loop(cond, step, (C_1, C_0, jnp.array(2, jnp.float64)))[0],
    )


@custom_vjp
def gegenbauer(level, alpha, x):
    """Gegenbauer function with custom gradient."""
    return _gegenbauer(level, alpha, x)


# @jax.jit
def gegenbauer_fwd(level, alpha, x):
    """The forward-mode differentiation for the Gegenbauer polynomial wrt the input `x`."""
    geg = _gegenbauer(level, alpha, x)
    g_fn = lambda: 2 * alpha * _gegenbauer(level - 1, alpha + 1, x)
    grad = jax.lax.cond(level == 0, lambda: jnp.zeros_like(x), g_fn)
    return geg, grad


# @jax.jit
def gegenbauer_bwd(res, cotangents):
    """The backward pass for the custom gradient of Gegenbauer polynomial wrt the input `x`."""
    return (None, None, res * cotangents)


# declare the custom vjp
gegenbauer.defvjp(gegenbauer_fwd, gegenbauer_bwd)


@dataclass
class GegenbauerLookupTable:
    """
    Gegenbauer polynomial via interpolation from a lookup table.

    The Gegenbauer polynomial are defined on the interval [-1, 1]. So we first evaluate offline on a
    dense grid the gegenbauer and its gradient on the interval. Then during training we interpolate
    from the lookup table.

    Attributes:
        max_level: The maximum order of the polynomial up to which we will evaluate on a grid.
        alpha: The hyper-sphere constant given by (d - 2) / 2 for the Sᵈ⁻¹ sphere.
        grid_resolution: The number of points in the linspace of [-1, 1]. Defautls to `100_000`.
        grid_evaluations: The grid evaluation of all orders of the Gegenbauer.
        grid_grad_evaluations: The grid evaluation of all orders of the gradient of Gegenbauer.
    """

    max_level: int = field(default_factory=int, pytree_node=False)
    alpha: float = field(default_factory=float, pytree_node=False)
    grid_resolution: int = field(default=100_000, pytree_node=False)
    grid_evaluations: jax.Array = field(init=False, pytree_node=False)
    grid_grad_evaluations: jax.Array = field(init=False, pytree_node=False)

    def __post_init__(self):
        """Evaluate the Gegenbauer and the gradients and save it to the relevant lookup tables."""
        x_grid = jnp.linspace(-1, 1, self.grid_resolution)

        # grid evaluation for the Gegenbauer
        part_gegenbauer = Partial(_gegenbauer, alpha=self.alpha, x=x_grid)
        grid_evaluations = jax.vmap(part_gegenbauer)(
            jnp.arange(self.max_level + 1, dtype=jnp.int32)
        )

        # grid evaluation for the gradient of Gegenbauer
        part_gegenbauer = Partial(_gegenbauer, alpha=self.alpha + 1, x=x_grid)
        grid_grad_evaluations = (
            2 * self.alpha * jax.vmap(part_gegenbauer)(jnp.arange(self.max_level, dtype=jnp.int32))
        )
        grid_grad_evaluations = jnp.stack([jnp.zeros_like(x_grid), *grid_grad_evaluations])
        grid_grad_evaluations = grid_grad_evaluations

        object.__setattr__(self, "grid_evaluations", grid_evaluations)
        object.__setattr__(self, "grid_grad_evaluations", grid_grad_evaluations)

    def __call__(self, level: int, alpha: Union[float, Float[Array, ""]], x: Array) -> Array:
        """
        Compute the Gegenbauer polynomial via linear interpolation from the lookup table.

        NOTE: The `alpha` argument is ignored as the Gegenbauer has been already computed with a
        user defined alpha.

        Args:
            level: The order of the polynomial.
            alpha: The hyper-sphere constant given by (d - 2) / 2 for the Sᵈ⁻¹ sphere.
            x: Input array.

        Returns:
            The interpolated Gegenbauer polynomial evaluated at the input `x`.
        """
        return gegenbauer_from_lookup(level, self.grid_evaluations, self.grid_grad_evaluations, x)


@custom_vjp
@jax.jit
def gegenbauer_from_lookup(
    level: float, grid_eval: Float[Array, "N M"], grid_grad_eval: Float[Array, "N M"], x: Array
) -> Array:
    """
    Evaluate the Gegenbauer polynomial and the custom gradient via interpoloation on a grid.

    Args:
        level: The order of the polynomial.
        grid_eval: The lookup table with the Gegenbauer evaluation.
        grid_grad_eval: The lookup table with the gradient of Gegenbauer.
        x: Input array.

    Returns:
        The interpolated Gegenbauer polynomial evaluated at the input `x`.
    """
    return tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grid_eval[jnp.array(level, jnp.int32)])


def gegenbauer_from_lookup_fwd(
    level: float, grid_eval: Float[Array, "N M"], grid_grad_eval: Float[Array, "N M"], x: Array
) -> Tuple[Array, Array]:
    """The forward-mode custom gradient wrt the input `x`."""
    return (
        gegenbauer_from_lookup(level, grid_eval, grid_grad_eval, x),
        tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grid_grad_eval[jnp.array(level, jnp.int32)]),
    )


def gegenbauer_from_lookup_bwd(res, cotangents):
    """The backward pass for the custom gradient wrt the input `x`."""
    return (None, None, None, res * cotangents)


# declare the custom vjp
gegenbauer_from_lookup.defvjp(gegenbauer_from_lookup_fwd, gegenbauer_from_lookup_bwd)


# @jax.custom_jvp
# def _evaluate2(level: int, grid_eval: Array, grid_grad_eval: Array, x: Array):
#     return tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grid_eval[int(level)])
#
#
# @_evaluate2.defjvp
# def _evaluate2_jvp(primals, tangents):
#     level, grid_eval, grad_eval, x = primals
#     *_, x_dot = tangents
#     dC_dx = tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grad_eval[int(level)])
#
#     primal_out = _evaluate2(*primals)
#     return primal_out, dC_dx * x_dot
#
#
# @jax.custom_gradient
# def _evaluate3(level: float, grid_eval: Array, grid_grad_eval: Array, x: Array):
#     def grad(g):
#         dC_dx = tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grid_grad_eval[int(level)])
#         return (None, None, None, dC_dx * g)
#
#     return tfp.math.interp_regular_1d_grid(x, -1.0, 1.0, grid_eval[int(level)]), grad



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from scipy.special import gegenbauer as sci_gegenbauer

from gpfy.gegenbauer import GegenbauerLookupTable, gegenbauer

key = jr.PRNGKey(42)


def _get_input_from_shape(shape):
    return jr.uniform(key, shape, minval=-1.0, maxval=1.0, dtype=jnp.float64)


alpha_list = [0.5, 1.5, 3.0, 10.0]
freq_list = [0, 1, 2, 5, 10]
shape_list = [(), (1, 1), (3, 1), (100, 7)]


@pytest.mark.parametrize("n", freq_list)
@pytest.mark.parametrize("alpha", alpha_list)
@pytest.mark.parametrize("shape", shape_list)
def test_gegenbauer_against_scipy(n, alpha, shape):
    """test gegenbauer while-loop vs scipy."""
    angle = _get_input_from_shape(shape)
    sci_geg = sci_gegenbauer(n, alpha)
    jnp.array_equal(sci_geg(angle), gegenbauer(n, alpha, angle))


@pytest.mark.parametrize("n", freq_list)
@pytest.mark.parametrize("alpha", alpha_list)
@pytest.mark.parametrize("shape", shape_list)
def test_gegenbauer_lookup_table_against_while_loop(n, alpha, shape):
    """test gegenbauer lookup table vs while-loop."""
    angle = _get_input_from_shape(shape)
    geg = GegenbauerLookupTable(max_level=n, alpha=alpha, grid_resolution=100_000)
    np.testing.assert_allclose(geg(n, alpha, angle), gegenbauer(n, alpha, angle), rtol=1e-5)


@pytest.mark.parametrize("n", freq_list)
@pytest.mark.parametrize("alpha", alpha_list)
def test_gegenbauer_gradient(n, alpha):
    """test gegenbauer gradient."""
    angle = _get_input_from_shape(shape=())
    geg = GegenbauerLookupTable(max_level=n, alpha=alpha, grid_resolution=100_000)
    lookup_grad = jax.grad(geg, argnums=2)(n, alpha, angle)
    gegenbauer_grad = jax.grad(gegenbauer, argnums=2)(n, alpha, angle)
    np.testing.assert_allclose(lookup_grad, gegenbauer_grad, rtol=1e-5)


def test_raises_when_initialise_lookuptable_with_grid_evaluations():
    """test we can't initialise grid_evaluations for LookupTable"""
    with pytest.raises(TypeError):
        GegenbauerLookupTable(2, 0.5, grid_evaluations=42)


def test_raises_when_initialise_lookuptable_with_grid_grad_evaluations():
    """test we can't initialise grid_grad_evaluations for LookupTable"""
    with pytest.raises(TypeError):
        GegenbauerLookupTable(2, 0.5, grid_grad_evaluations=42)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Borrowed and adapted from the excellent implementation of GPflow (and later on of GPJax)


from typing import Callable

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float

"""The number of Gauss-Hermite points to use for quadrature"""
DEFAULT_NUM_GAUSS_HERMITE_POINTS = 20
gh_points, gh_weights = np.polynomial.hermite.hermgauss(DEFAULT_NUM_GAUSS_HERMITE_POINTS)


def gauss_hermite_quadrature(
    fun: Callable[[Float[Array, "N ... D"], Float[Array, "N ... D"]], Float[Array, "N ..."]],
    mean: Float[Array, "N D"],
    stddev: Float[Array, "N D"],
    *args,
    **kwargs,
) -> Callable[[Float[Array, "N D"]], Float[Array, " N"]]:
    """
    Compute Gaussian-Hermite quadrature for a given function. The quadrature
    points are adjusted through the supplied mean and variance arrays.
    Args:
        fun (Callable): The function for which quadrature should be applied to.
        mean (Float[Array, "N D"]): The mean of the Gaussian distribution that
            is used to shift quadrature points.
        sd (Float[Array, "N D"]): The standard deviation of the Gaussian
            distribution that is used to scale quadrature points.
    Returns:
        Float[Array, "N"]: The evaluated integrals value.
    """
    # [N, num_gh, D]

    X = mean[..., None, :] + jnp.sqrt(2.0) * stddev[..., None, :] * gh_points[..., None]
    W = gh_weights / jnp.sqrt(jnp.pi)  # [gh,]
    return lambda y: jnp.sum(fun(X, y[..., None, :]) * W, axis=-1)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from jax.tree_util import tree_leaves, tree_map

from gpfy.spherical_harmonics import SphericalHarmonics, _num_harmonics

key = jr.PRNGKey(42)


freq_list = [2, 5]
phases_list = [5, 10]


def test_init_fails_when_passing_params_with_wrong_sphere(kernel_param_5d):
    sh = SphericalHarmonics(5, 10)
    kernel_param = kernel_param_5d[1]
    with pytest.raises(ValueError):
        _ = sh.init(key, input_dim=4, param=kernel_param)


@pytest.mark.parametrize("use_kernel_param", [True, False])
@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_init_create_correct_variables(use_kernel_param, freq, phases, kernel_param_5d):
    input_dim = 5
    kernel_param = kernel_param_5d[1]
    kernel_param = kernel_param if use_kernel_param else None
    sh = SphericalHarmonics(freq, phases)
    param = sh.init(key, input_dim=input_dim, param=kernel_param)

    assert all(
        f"V_{f:02d}" in param.params["variational"]["inducing_features"] for f in range(freq)
    )

    # check trainability
    for n, v in enumerate(param._trainables["variational"]["inducing_features"].values()):
        assert v is False if phases >= _num_harmonics(input_dim + 1, n) else True

    assert "sphere" in param.constants
    assert "gegenbauer_lookup_table" in param.constants["sphere"]

    orth_basis = param.constants["variational"]["inducing_features"]["orthogonal_basis"]
    if any(tree_leaves(param._trainables["variational"]["inducing_features"])):
        assert all(tree_map(lambda b: jnp.isnan(b).all(), orth_basis))
    else:
        assert not any(tree_map(lambda b: jnp.isnan(b).any(), orth_basis))


@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_correct_levels(freq, phases):
    sh = SphericalHarmonics(freq, phases)
    np.testing.assert_allclose(sh.levels, jnp.arange(freq))


@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_Vs_are_normalised(freq, phases):
    input_dim = 4
    sh = SphericalHarmonics(freq, phases)
    param = sh.init(key, input_dim=input_dim)
    Vs = sh.Vs(param)
    norms = tree_map(lambda v: jax.vmap(lambda vv: jnp.dot(vv, vv))(v), Vs)
    np.testing.assert_allclose(jnp.concatenate(norms), 1.0)


def test_num_phases_do_not_exceed_truncation_level():
    input_dim = 5
    phase_truncation = 3  # from the 2nd frequency we have to truncat since 2nd has num_dim phases.
    sh = SphericalHarmonics(3, phase_truncation=phase_truncation)
    param = sh.init(key, input_dim=input_dim)

    assert all(phases <= phase_truncation for phases in sh.num_phase_in_frequency(param))


@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_num_inducing_matches_shape_of_Vs(freq, phases):
    input_dim = 4
    sh = SphericalHarmonics(freq, phases)
    param = sh.init(key, input_dim=input_dim)
    assert jnp.concatenate(sh.Vs(param), 0).shape[0] == sh.num_inducing(param)


@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_Kuu_shape(freq, phases, kernel_param_5d):
    k, kernel_param = kernel_param_5d
    input_dim = 5
    sh = SphericalHarmonics(freq, phases)
    param = sh.init(key, input_dim=input_dim, param=kernel_param)
    Kuu = sh.Kuu(param, k)
    assert Kuu.shape[0] == sh.num_inducing(param)


@pytest.mark.parametrize("freq", freq_list)
@pytest.mark.parametrize("phases", phases_list)
def test_Kuf_shape(freq, phases, kernel_param_5d):
    k, kernel_param = kernel_param_5d
    input_dim = 5
    x = jr.normal(key, (10, 5), dtype=jnp.float64)
    sh = SphericalHarmonics(freq, phases)
    param = sh.init(key, input_dim=input_dim, param=kernel_param)
    Kuf = sh.Kuf(param, k, x)
    np.testing.assert_array_equal(Kuf.shape, (sh.num_inducing(param), x.shape[0]))



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from scipy.integrate import quad

from gpfy.gegenbauer import gegenbauer
from gpfy.harmonics.utils import weight_func

alpha_list = [0.5, 1.5, 3.0, 10.0]
freq_list = [0, 1, 2, 5, 10]


@pytest.mark.parametrize("n", freq_list)
@pytest.mark.parametrize("alpha", alpha_list)
def test_orthogonality_of_the_gegenbauer_under_the_weight_function(n, alpha):
    """Check the Gegenbauer orthogonality wrt to the weight function."""
    dim = int(2 * alpha + 2)

    def integrand(x):
        return gegenbauer(n, alpha, x) * gegenbauer(n + 1, alpha, x) * weight_func(x, dim)

    res, tol = quad(integrand, -1, 1)

    np.testing.assert_allclose(res, 0, atol=tol)



# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# This is a copy from the excellent flax.struct.py with minor modifications. Mainly to remove
# serialization for now


import dataclasses
from typing import TypeVar

import jax
from typing_extensions import dataclass_transform

jax.config.update("jax_enable_x64", True)
_T = TypeVar("_T")


def field(pytree_node=True, **kwargs):
    return dataclasses.field(metadata={"pytree_node": pytree_node}, **kwargs)


@dataclass_transform(field_specifiers=(field,))  # type: ignore[literal-required]
def dataclass(clz: _T) -> _T:
    """Create a class which can be passed to functional transformations.
    NOTE: Inherit from ``PyTreeNode`` instead to avoid type checking issues when
    using PyType.
    Jax transformations such as `jax.jit` and `jax.grad` require objects that are
    immutable and can be mapped over using the `jax.tree_util` methods.
    The `dataclass` decorator makes it easy to define custom classes that can be
    passed safely to Jax. For example::
      from flax import struct
      @struct.dataclass
      class Model:
        params: Any
        # use pytree_node=False to indicate an attribute should not be touched
        # by Jax transformations.
        apply_fn: FunctionType = struct.field(pytree_node=False)
        def __apply__(self, *args):
          return self.apply_fn(*args)
      model = Model(params, apply_fn)
      model.params = params_b  # Model is immutable. This will raise an error.
      model_b = model.replace(params=params_b)  # Use the replace method instead.
      # This class can now be used safely in Jax to compute gradients w.r.t. the
      # parameters.
      model = Model(params, apply_fn)
      model_grad = jax.grad(some_loss_fn)(model)
    Note that dataclasses have an auto-generated ``__init__`` where
    the arguments of the constructor and the attributes of the created
    instance match 1:1. This correspondence is what makes these objects
    valid containers that work with JAX transformations and
    more generally the `jax.tree_util` library.
    Sometimes a "smart constructor" is desired, for example because
    some of the attributes can be (optionally) derived from others.
    The way to do this with Flax dataclasses is to make a static or
    class method that provides the smart constructor.
    This way the simple constructor used by `jax.tree_util` is
    preserved. Consider the following example::
      @struct.dataclass
      class DirectionAndScaleKernel:
        direction: Array
        scale: Array
        @classmethod
        def create(cls, kernel):
          scale = jax.numpy.linalg.norm(kernel, axis=0, keepdims=True)
          directin = direction / scale
          return cls(direction, scale)
    Args:
      clz: the class that will be transformed by the decorator.
    Returns:
      The new class.
    """
    # check if already a flax dataclass
    if "_flax_dataclass" in clz.__dict__:
        return clz

    data_clz = dataclasses.dataclass(frozen=True)(clz)  # type: ignore
    meta_fields = []
    data_fields = []
    for field_info in dataclasses.fields(data_clz):
        is_pytree_node = field_info.metadata.get("pytree_node", True)
        if is_pytree_node:
            data_fields.append(field_info.name)
        else:
            meta_fields.append(field_info.name)

    def replace(self, **updates) -> _T:
        """ "Returns a new object replacing the specified fields with new values."""
        return dataclasses.replace(self, **updates)

    # data_clz.replace = replace  # type: ignore
    setattr(data_clz, "replace", replace)

    def iterate_clz(x):
        meta = tuple(getattr(x, name) for name in meta_fields)
        data = tuple(getattr(x, name) for name in data_fields)
        return data, meta

    def iterate_clz_with_keys(x):
        meta = tuple(getattr(x, name) for name in meta_fields)
        data = tuple((jax.tree_util.GetAttrKey(name), getattr(x, name)) for name in data_fields)
        return data, meta

    def clz_from_iterable(meta, data):
        meta_args = tuple(zip(meta_fields, meta))
        data_args = tuple(zip(data_fields, data))
        kwargs = dict(meta_args + data_args)
        return data_clz(**kwargs)

    jax.tree_util.register_pytree_with_keys(data_clz, iterate_clz_with_keys, clz_from_iterable)

    # add a _flax_dataclass flag to distinguish from regular dataclasses
    data_clz._flax_dataclass = True  # type: ignore[attr-defined]

    return data_clz  # type: ignore


TNode = TypeVar("TNode", bound="PyTreeNode")


@dataclass_transform(field_descriptors=(field,))  # type: ignore[literal-required]
class PyTreeNode:
    """Base class for dataclasses that should act like a JAX pytree node.
    See ``flax.struct.dataclass`` for the ``jax.tree_util`` behavior.
    This base class additionally avoids type checking errors when using PyType.
    Example::
      from flax import struct
      class Model(struct.PyTreeNode):
        params: Any
        # use pytree_node=False to indicate an attribute should not be touched
        # by Jax transformations.
        apply_fn: FunctionType = struct.field(pytree_node=False)
        def __apply__(self, *args):
          return self.apply_fn(*args)
      model = Model(params, apply_fn)
      model.params = params_b  # Model is immutable. This will raise an error.
      model_b = model.replace(params=params_b)  # Use the replace method instead.
      # This class can now be used safely in Jax to compute gradients w.r.t. the
      # parameters.
      model = Model(params, apply_fn)
      model_grad = jax.grad(some_loss_fn)(model)
    """

    def __init_subclass__(cls):
        dataclass(cls)  # pytype: disable=wrong-arg-types

    def __init__(self, *args, **kwargs):
        # stub for pytype
        raise NotImplementedError

    def replace(self: TNode, **overrides) -> TNode:
        # stub for pytype
        raise NotImplementedError



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow_probability.substrates.jax as tfp

from gpfy.param import Param, identity, positive


def test_param_raises_with_invalid_tree_structure():
    """Test we raise if the structure does not much."""
    with pytest.raises(ValueError):
        Param(params={"a": 0}, _trainables={"b": False})

    with pytest.raises(ValueError):
        Param(params={"a": {"b": 0}}, _trainables={"c": 3.0})

    with pytest.raises(ValueError):
        Param(params={"a": 0}, _bijectors={"b": identity()})

    with pytest.raises(ValueError):
        Param(params={"a": {"b": 0}}, _bijectors={"c": positive()})

    with pytest.raises(ValueError):
        Param(params={"a": 0}, constants={"b": 3.0})


def test_is_subtree():
    """Test we correctly detect sub-trees starting from same level."""
    dummy = Param()
    assert dummy._is_subtree({"a": {"b": False}}, {"a": {"b": 1.0, "c": 0.0}})
    assert dummy._is_subtree({"b": False}, {"b": 1.0, "c": 0.0})
    assert dummy._is_subtree({"b": False}, {"b": 1.0})
    assert not dummy._is_subtree(1.0, {"c": 1.0})
    assert not dummy._is_subtree({"b": False}, {"c": 1.0})
    assert not dummy._is_subtree({"c": False}, {"a": {"b": 1.0}})


def test_param_construction_can_correctly_fill_the_dictionaries():
    """Test we correctly initialise bijectors and trainable status for unspecified variables."""
    param = Param(params={"a": 0.0, "b": 1.0}, _trainables={"b": False})
    assert jax.tree_util.tree_structure(param.params) == jax.tree_util.tree_structure(
        param._trainables
    )
    assert param._trainables["a"] is True  # by default we have trainables=True
    assert param._trainables["b"] is False

    assert jax.tree_util.tree_structure(param.params) == jax.tree_util.tree_structure(
        param._bijectors, is_leaf=lambda x: isinstance(x, tfp.bijectors.Bijector)
    )

    # default bijector is positive
    assert isinstance(param._bijectors["a"], positive)
    assert isinstance(param._bijectors["b"], positive)


def test_replace_param():
    """Test we can replace the params."""
    param = Param(params={"collection_a": {"a": 0.0, "b": 1.0}, "collection_b": {"b": 2.0}})
    new_param = param.replace_param(collection="collection_a", b=42.0)
    assert new_param.params["collection_a"]["b"] == 42.0
    assert new_param.params["collection_b"]["b"] == 2.0
    new_param = new_param.replace_param(collection="collection_b", b=42.0)
    assert new_param.params["collection_b"]["b"] == 42.0

    with pytest.raises(ValueError):
        _ = param.replace_param(collection="unknown_collection", b=3.0)


def test_set_trainable():
    """Test we can change trainable status."""
    param = Param(params={"collection_a": {"a": 0.0, "b": 1.0}, "collection_b": {"b": 2.0}})
    assert all(v is True for v in param._trainables["collection_a"].values())
    assert all(v is True for v in param._trainables["collection_b"].values())

    param = param.set_trainable("collection_a", a=False, b=False)
    assert all(v is False for v in param._trainables["collection_a"].values())
    assert param._trainables["collection_b"]["b"]
    param = param.set_trainable("collection_b", b=False)
    assert all(v is False for v in param._trainables["collection_b"].values())

    with pytest.raises(ValueError):
        _ = param.set_trainable(collection="unknown_collection", b=False)


def test_set_bijector():
    """Test we can set bijector."""
    param = Param(params={"collection_a": {"a": 0.0, "b": 1.0}, "collection_b": {"b": 2.0}})
    assert all(isinstance(v, positive) for v in param._bijectors["collection_a"].values())
    assert all(isinstance(v, positive) for v in param._bijectors["collection_b"].values())

    param = param.set_bijector("collection_a", a=identity(), b=identity())
    assert all(isinstance(v, identity) for v in param._bijectors["collection_a"].values())
    assert isinstance(param._bijectors["collection_b"]["b"], positive)
    param = param.set_bijector("collection_b", b=identity())
    assert all(isinstance(v, identity) for v in param._bijectors["collection_b"].values())

    with pytest.raises(ValueError):
        _ = param.set_bijector(collection="unknown_collection", b=identity())


def test_constrained_and_unconstrained():
    """Test we can apply fwd-inverse bijector."""
    param = Param(
        params={"collection_a": {"a": 0.1, "b": 1.0}, "collection_b": {"c": 2.0}},
        _bijectors={
            "collection_a": {"a": positive(), "b": identity()},
            "collection_b": {"c": positive()},
        },
    )

    unconstrained = param.unconstrained()

    expected_params = {
        "collection_a": {"a": jnp.log(0.1).astype(jnp.float64), "b": 1.0},
        "collection_b": {"c": jnp.log(2.0).astype(jnp.float64)},
    }

    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_allclose, unconstrained.params, expected_params
    )

    constrained = unconstrained.constrained()
    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_allclose, constrained.params, param.params
    )



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Tuple

import jax
import jax.numpy as jnp
from jax._src.lax.linalg import triangular_solve
from jax.tree_util import tree_leaves, tree_map
from jaxtyping import Array, Float
from scipy.special import comb

from gpfy.gegenbauer import GegenbauerLookupTable, gegenbauer
from gpfy.harmonics.fund_set import fundamental_set_loader
from gpfy.param import Param, identity
from gpfy.spherical import Spherical
from gpfy.typing import PRNG, BijectorDict, ConstantDict, TrainableDict, VariableDict
from gpfy.utils import dataclass, field


def _num_harmonics(dim: int, frequency: int) -> int:
    """
    Compute the number of spherical harmonic functions in a given frequency.

    Args:
        dim: The dimensionality of the input on the Sᵈ⁻¹ sphere.
        frequency: The order of the frequency.

    Returns:
        The number of spherical harmonic functions for the the nth frequency.
    """
    if frequency == 0:
        return 1
    else:
        c1 = comb(frequency + dim - 3, frequency - 1)
        c2 = comb(frequency + dim - 2, frequency)
    return int(c1 + c2)


@dataclass
class SphericalHarmonics:
    """
    Spherical harmonics inducing features for sparse inference in Gaussian processes.

    The spherical harmonics, Yₙᵐ(·) of frequency n and phase m are eigenfunctions on the sphere and,
    as such, they form an orthogonal basis.

    To construct the harmonics, we use a a fundamental set of points on the sphere {vᵢ}ᵢ and compute
    b = {Cᵅₙ(<vᵢ, x>)}ᵢ. b now forms a complete basis on the sphere and we can orthogoalise it via
    a Cholesky decomposition. However, we only need to run the Cholesky decomposition once during
    initialisation.

    NOTE: If we truncate the harmonics up to a maximum phase, then we need to learn the fundamental
    set of points {vᵢ}ᵢ from data. This requires to run a Cholesky decomposition at every iteration.

    Attributes:
        num_frequencies: The number of frequencies, up to which, we compute the harmonics.
        phase_truncation: The number for truncating the phases/harmonics in every frequency.
            Defaults to max(int), which essentially means no truncation.

    Returns:
        An instance of the spherical harmonics features.
    """

    num_frequencies: int = field(default_factory=10, pytree_node=False)
    phase_truncation: int = field(default=2**31 - 1, pytree_node=False)

    @property
    def levels(self):
        return jnp.arange(self.num_frequencies, dtype=jnp.int32)

    def init(self, key: PRNG, input_dim: int, param: Optional[Param] = None) -> Param:
        """
        Initialise the parameters of the spherical harmonic features and return a `Param` object.

        Args:
            key: A random key for initialising the fundamental set in case we truncate the phases.
            input_dim: The input dimension. We then append an extra dimension for bias.
            param: An already initialised collection of parameters from another object, i.e., a
                kernel. If it is provided we extend the `Param` with the additional collection.
                Otherwise we return a new `Param`. Defaults to None.

        Raises:
            ValueError: if a `param` is provided but the `sphere_dim`s do not match.

        Returns:
            The `Param` object with all variables.
        """
        sphere_dim = input_dim
        dim = sphere_dim + 1

        # check if we have a sphere initialised in our param object and fail if it doesn't match.
        sphere = param.constants.get("sphere") if param else {}
        if sphere and sphere_dim != sphere["sphere_dim"]:
            raise ValueError("`param` contains a sphere that is not compatible to `input_dim`.")

        # Try loading a pre-computed fundamental set.
        try:
            fund_set = fundamental_set_loader(dim)
        except ValueError:
            fund_set = None

        # initialise the parameters Vs. Set them to non-trainable if we do not truncate the phase.
        Vs = {}
        trainables = {}
        for n in self.levels:
            num_phase = _num_harmonics(dim, n)

            if (num_phase <= self.phase_truncation) and fund_set:
                V = fund_set(n)
                Vs[f"V_{n:02d}"] = V
                trainables[f"V_{n:02d}"] = False
            else:
                key, subkey = jax.random.split(key)
                V = jax.random.normal(subkey, (min(self.phase_truncation, num_phase), dim))
                Vs[f"V_{n:02d}"] = V
                trainables[f"V_{n:02d}"] = True

        # initialise the orthogonal basis with nans to get the structure
        orth_basis = tree_leaves(
            tree_map(lambda x: jnp.nan * jnp.zeros((x.shape[0], x.shape[0]), dtype=x.dtype), Vs)
        )

        # create the collections for the current object.
        bijectors = {k: identity() for k in Vs.keys()}
        collection = "variational"
        var_params = {"inducing_features": Vs}
        var_trainables = {"inducing_features": trainables}
        var_bijectors = {"inducing_features": bijectors}
        var_constants = {"inducing_features": {"orthogonal_basis": orth_basis}}

        # initialise the Gegenbauer lookup table and compute the relevant constatns on the sphere.
        if (
            not sphere
            or ("gegenbauer_lookup_table" not in sphere)
            or (sphere["gegenbauer_lookup_table"].max_level < self.num_frequencies)
        ):
            sphere = sphere or {}
            alpha = (dim - 2.0) / 2.0
            sphere["sphere_dim"] = sphere_dim
            sphere["alpha"] = alpha
            geg = GegenbauerLookupTable(self.num_frequencies, alpha)
            sphere["gegenbauer_lookup_table"] = geg

        all_params: VariableDict = {}
        all_trainables: TrainableDict = {}
        all_bijectors: BijectorDict = {}
        all_constants: ConstantDict = {}
        constrained = True

        # get the collections in case we have a provided Param
        if param:
            all_params = param.params
            all_bijectors = param._bijectors
            all_trainables = param._trainables
            all_constants = param.constants
            constrained = param._constrained

        # append the initialised parameters for spherical harmonic features to the collections
        all_params[collection] = var_params
        all_bijectors[collection] = var_bijectors
        all_trainables[collection] = var_trainables
        all_constants[collection] = var_constants
        all_constants["sphere"] = sphere

        # create a new Param object.
        param = Param(
            params=all_params,
            _trainables=all_trainables,
            _bijectors=all_bijectors,
            constants=all_constants,
            _constrained=constrained,
        )

        # if we don't have any phase_truncation (so no trainable vars), pre-compute and save the
        # orthogonal basis, otherwise keep it with NaNs.
        if not any(tree_leaves(var_trainables)):
            orth_basis = self.orthogonalise_basis(param)

        all_constants[collection] = {"inducing_features": {"orthogonal_basis": orth_basis}}
        param = param.replace(constants=all_constants)
        return param

    def Vs(self, param: Param) -> List[Array]:
        """
        Get the fundamental set of points.

        NOTE: if we learn the fundamental set, then we normalise the points.

        Args:
            param: A `Param` initialised with the spherical harmonic features.

        Returns:
            A list containing the fundamental set of points at every frequency.
        """
        Vs = tree_leaves(param.params["variational"]["inducing_features"])

        # the orth_basis is nans if we learn the Vs
        orth_basis = param.constants["variational"]["inducing_features"]["orthogonal_basis"]

        def _normalise():
            return tree_map(
                lambda v: v / jnp.sqrt(jnp.sum(jnp.square(v), axis=1, keepdims=True)),
                Vs,
            )

        return jax.lax.cond(jnp.isnan(jnp.sum(orth_basis[0])), _normalise, lambda: Vs)

    def Ls(self, param: Param) -> List[Array]:
        """
        Compute the orthogonalised basis.

        NOTE: We compute the orthogonalisation only once, at the initialisation, if we don't
        truncate the phases/harmonics. Otherwise we need to orthogonalise at every iteration.

        Args:
            param: A `Param` initialised with the spherical harmonic features.

        Returns:
            A list containing the orthogonal basis at every frequency.
        """
        # the orth_basis is nans if we learn the Vs
        orth_basis = param.constants["variational"]["inducing_features"]["orthogonal_basis"]

        return jax.lax.cond(
            jnp.isnan(jnp.sum(orth_basis[0])),
            lambda: self.orthogonalise_basis(param),
            lambda: orth_basis,
        )

    def num_phase_in_frequency(self, param: Param) -> List[int]:
        """
        Get the total number of phases/harmonics at every frequency.

        Args:
            param: A `Param` initialised with the spherical harmonic features.

        Returns:
            A list with the number of phases per frequency.
        """
        return jax.tree_map(lambda x: x.shape[0], self.Vs(param))

    def num_inducing(self, param) -> int:
        """
        Computes the total number of inducing features, as the sum of all phases.

        Args:
            param: A `Param` initialised with the spherical harmonic features.

        Returns:
            The total number of inducing features.
        """
        return sum(self.num_phase_in_frequency(param))

    def orthogonalise_basis(self, param: Param) -> List[Array]:
        """
        Compute the basis from the fundamental set and orthogonalise it via Cholesky decomposition.

        Args:
            param: A `Param` initialised with the spherical harmonic features.

        Returns:
            A list containing the orthogonal basis at every frequency.
        """
        alpha = param.constants["sphere"]["alpha"]
        # gegenbauer = param.constants["sphere"]["gegenbauer_lookup_table"]
        levels = jnp.split(self.levels, self.num_frequencies)
        const = alpha / (alpha + self.levels.astype(jnp.float64))
        const = jnp.split(const, self.num_frequencies)

        def _func(v, n, c):
            x = jnp.matmul(v, v.T)
            B = c * gegenbauer(n[0], alpha, x)
            L = jnp.linalg.cholesky(B + 1e-16 * jnp.eye(B.shape[0], dtype=B.dtype))
            return L

        Ls = tree_map(_func, self.Vs(param), levels, const)
        return Ls

    def polynomial_expansion(self, param: Param, X: Float[Array, "N D"]) -> Float[Array, "M N"]:
        """
        Evaluate the polynomial expansion of an input on the sphere given the harmonic basis.

        Args:
            param: A `Param` initialised with the spherical harmonic features.
            X: Input Array.

        Returns:
            The harmonics evaluated at the input as a polynomial expansion of the basis.
        """
        alpha = param.constants["sphere"]["alpha"]
        # gegenbauer = param.constants["sphere"]["gegenbauer_lookup_table"]
        levels = jnp.split(self.levels, self.num_frequencies)
        const = alpha / (alpha + self.levels.astype(jnp.float64))
        const = jnp.split(const, self.num_frequencies)

        def _func(v, n, L):  # , c):
            vxT = jnp.dot(v, X.T)
            zonal = gegenbauer(n[0], alpha, vxT)

            # vvT = jnp.matmul(v, v.T)
            # B = c * gegenbauer(n[0], alpha, vvT)
            # # B = c * geg2(n[0], alpha, vvT)
            # L = jnp.linalg.cholesky(B + 1e-16 * jnp.eye(B.shape[0], dtype=B.dtype))

            harmonic = triangular_solve(L, zonal, left_side=True, lower=True)
            return harmonic

        harmonics = tree_map(_func, self.Vs(param), levels, self.Ls(param))  # , const)
        return jnp.concatenate(harmonics, axis=0)

    def Kuu(self, param: Param, kernel: Spherical) -> Float[Array, " M"]:
        """
        Compute the covariance between the harmonic features.

        It is as a diagonal matrix which holds the reciprocal of the eigenvalues.

        Args:
            param: A `Param` initialised with the spherical harmonic features.
            kernel: A `Spherical` kernel.

        Returns:
            The covariance of the inducing features.
        """
        eigs = kernel.eigenvalues(param, self.levels)

        # split the eigenvalue array to a list so we can pass it to tree_map.
        eigs = jnp.split(eigs, self.num_frequencies)

        # get the repetition of each eigenvalue, i.e., the harmonic phases in each frequency.
        reps = self.num_phase_in_frequency(param)
        return jnp.concatenate(jax.tree_util.tree_map(lambda e, r: jnp.ones(r) / e, eigs, reps))

    def Kuf(self, param: Param, kernel: Spherical, x: Float[Array, "N D"]) -> Float[Array, "M N"]:
        """
        Compute the covariance between the harmonic features and the provided input.

        Args:
            param: A `Param` initialised with the spherical harmonic features.
            kernel: A `Spherical` kernel.
            x: Input Array.

        Returns:
            The cross covariance between the inducing features and the input.
        """
        # project the input to the unit sphere.
        x_sphere, rx = kernel.to_sphere(param, x)

        # get the harmonics evaluated at the input as a polynomial expansion of the basis.
        sh = self.polynomial_expansion(param, x_sphere)
        variance = param.params[kernel.name]["variance"]
        return sh * rx[..., None, :] * jnp.sqrt(variance)  # apply back to radius before returning

    def conditional_fun(
        self,
        kernel: Spherical,
    ) -> Tuple[
        Callable[[Param, Float[Array, "N D"]], Float[Array, "M N"]],
        Callable[[Param, Float[Array, "N D"], Float[Array, "M N"]], Float[Array, " N"]],
        Callable[[Param, Float[Array, "N D"], Float[Array, "M N"]], Float[Array, "N N"]],
    ]:
        """
        Prepare the necessary functions for conditioning a GP on the spherial harmonic features.

        We do not explicitly evalutate the functions here as we only want to pass it to a GP so that
        we know how to perform the conditioning.

        Args:
            kernel: The `Spherical` kernel of the GP.

        Returns:
            A tuple of 3 functions for computing the projection of the input, the conditional
            diagonal variance and the conditional covariance.
        """
        project_fun = lambda param, x: jnp.sqrt(1 / self.Kuu(param, kernel))[..., None] * self.Kuf(
            param, kernel, x
        )
        conditional_var_fun = lambda param, x, proj: kernel.K_diag(param, x) - jnp.sum(
            jnp.square(proj), -2
        )
        conditional_cov_fun = lambda param, x, proj: kernel.K(param, x) - jnp.matmul(
            proj.swapaxes(-1, -2), proj
        )

        return (project_fun, conditional_cov_fun, conditional_var_fun)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Callable

import numpy as np
from jax import Array


def fundamental_set_loader(dimension: int, load_dir="fundamental_system") -> Callable[[int], Array]:
    load_dir = Path(__file__).parents[0] / ".." / load_dir
    file_name = load_dir / f"fs_{dimension}D.npz"

    cache = {}
    if file_name.exists():
        with np.load(file_name) as f:
            cache = {k: v for (k, v) in f.items()}
    else:
        raise ValueError(f"Fundamental system for dimension {dimension} has not been precomputed.")

    def load(degree: int) -> Array:
        key = f"degree_{degree}"
        if key not in cache:
            raise ValueError()
        return cache[key]

    return load



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Union

import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from jax.tree_util import tree_map, tree_structure

from gpfy.typing import BijectorDict, ConstantDict, TrainableDict, VariableDict
from gpfy.utils import PyTreeNode, field

positive = tfp.bijectors.Exp
identity = tfp.bijectors.Identity


class Param(PyTreeNode):
    """
    Basic `PyTreeNode` that holds information regarding all the parameters of an initialised object.

    Attributes:
        params: A dictionary holding a collection of parameterised objects with a mapping to their
            parameters.
        _trainables: A dictionary with the same structure as `params` that specifies if a parameter
            is trainable or not. It defaults to `True` for all unspecified parameters.
        _bijectors: A dictionary with the same structure as `params` that specifies the required
            bijector to transform to the unconstrained space. It defaults to a positive
            `tfp.bijectors.Exp` for all unspecified parameters.
        constants: A dictionary that holds information for additional variables that are considered
            constant during optimisation.
        _constrained: A flag specifying if the parameters are constrained or not.

    Note that only the `params` attribute acts as a pytree_node.
    """

    params: VariableDict = field(default_factory=dict, pytree_node=True)
    _trainables: TrainableDict = field(default_factory=dict, pytree_node=False)
    _bijectors: BijectorDict = field(default_factory=dict, pytree_node=False)
    constants: ConstantDict = field(default_factory=dict, pytree_node=False)
    _constrained: bool = field(default=True, pytree_node=False)

    def _has_valid_keys(self) -> None:
        """
        Checks if the provided collections have the same structure as the `self.params`.

        Raises:
            ValueError: if the user specified `self._trainables is not a subtree of `self.params`.
            ValueError: if the user specified `self._bijectors is not a subtree of `self.params`.
            ValueError: if the user specified `self._constants has collections other than
                the ones in `self.params` or `"sphere"`.
        """
        # valid_keys = set(self.params.keys())
        # param_tree = jax.tree_util.tree_flatten_with_path(self.params)[1]
        # trainables_tree = jax.tree_util.tree_flatten_with_path(self._trainables)[1]
        # bijectors_tree = jax.tree_util.tree_flatten_with_path(self._bijectors)[1]
        if not self._is_subtree(self._trainables, self.params):
            raise ValueError("Invalid key in `_trainables`")
        if not self._is_subtree(self._bijectors, self.params):
            raise ValueError("Invalid key in `_bijectors`")
        if not all(
            collection in self.params or collection == "sphere"
            for collection in self.constants.keys()
        ):
            raise ValueError("Invalid key in `_constants`")

    def _is_subtree(self, t1: Union[VariableDict, Any], t2: Union[VariableDict, Any]) -> bool:
        """
        Check if `t1` is subtree of `t2`, strating from the same level.

        Args:
            t1: a `VariableDict` pytree or a leaf node
            t2: a `VariableDict` pytree or a leaf node

        Returns:
            If `t1` is a subtree of `t2`.
        """
        if isinstance(t1, Dict) and isinstance(t2, Dict):
            ret = []
            for k1 in t1.keys():
                if k1 in t2:  # Check if a subtree of t1 has same structure as in t2
                    ret.append(self._is_subtree(t1[k1], t2[k1]))
                else:
                    return False
            return all(ret)
        elif isinstance(t2, Dict):  # t1 is a leaf but t2 is a tree
            return False
        else:  # both t1 and t2 are leaves
            return True

    def _tree_update_from_subtree(self, t1: VariableDict, t2: VariableDict) -> VariableDict:
        """
        Update tree `t1` from subtree `t2`.

        Args:
            t1: a `VariableDict` pytree.
            t2: a `VariableDict` pytree.

        Returns:
            A `VariableDict` with the updated values.
        """
        ret = {}
        for k1, v1 in t1.items():
            if k1 in t2:  # k1 needs to be updated
                if isinstance(v1, Dict):  # v1 is a tree so recurse
                    ret[k1] = self._tree_update_from_subtree(t1[k1], t2[k1])
                else:
                    ret[k1] = t2[k1]  # we have a leaf so update the value
            else:
                if isinstance(v1, Dict):  # check if t2 is a subtree of t1 so we need to recurse
                    ret[k1] = self._tree_update_from_subtree(t1[k1], t2)
                else:
                    ret[k1] = v1  # no update needed
        return ret

    def __post_init__(self) -> None:
        """
        Runs automatically after the `__init__` of the dataclass to do further checks.
        """
        # check we have valid keys in all dicts
        self._has_valid_keys()

        # initialise the trainable status to `True` for all unpsecified variables
        trainables = self._trainables
        if not trainables or (tree_structure(trainables) != tree_structure(self.params)):
            trainables = tree_map(lambda _: True, self.params)
            trainables = self._tree_update_from_subtree(trainables, self._trainables)

        # initialising the bijectors to `positive` for all unpsecified variables
        bijectors = self._bijectors
        if not bijectors or (
            tree_structure(bijectors, is_leaf=lambda x: isinstance(x, tfp.bijectors.Bijector))
            != tree_structure(self.params)
        ):
            bijectors = tree_map(lambda _: positive(), self.params)
            bijectors = self._tree_update_from_subtree(bijectors, self._bijectors)

        # make sure all params are Arrays with float64 dtype
        params = tree_map(lambda x: jnp.array(x, dtype=jnp.float64), self.params)

        # write back the modified `VariableDict`s
        object.__setattr__(self, "params", params)
        object.__setattr__(self, "_trainables", trainables)
        object.__setattr__(self, "_bijectors", bijectors)

    def replace_param(self, collection: str, **kwargs) -> "Param":
        """
        Replace the value of parameters in the `VariableDict` from a specified `collection`.

        Args:
            collection: the name of the collection that holds the target variable.
            kwargs: The name and the new value of the target variables within the `collection`.

        Raises:
            ValueError: if the specified `collection` is not present in the `param` `VariableDict`.

        Returns:
            A new `Param` with the updated variables.
        """
        if collection not in self.params:
            raise ValueError(f"there is no {collection} collection")

        # first update the subtree within the specified collection
        updates = self._tree_update_from_subtree(self.params[collection], kwargs)
        # then update the params with the newly updated collection
        updates = self._tree_update_from_subtree(self.params, {collection: updates})
        return self.replace(params=updates)

    def set_trainable(self, collection: str, **kwargs) -> "Param":
        """
        Replace the trainable status of parameters in a specified `collection`.

        Args:
            collection: the name of the collection that holds the target variable.
            kwargs: The name and the new trainable status of the target variables within
                the `collection`.

        Raises:
            ValueError: if the specified `collection` is not present in the `param` `VariableDict`.

        Returns:
            A new `Param` with the updated trainable status of the variables.
        """
        if collection not in self._trainables:
            raise ValueError(f"there is no {collection} collection")

        # first update the subtree within the specified collection
        updates = self._tree_update_from_subtree(self._trainables[collection], kwargs)
        # then update the trainables with the newly updated collection
        updates = self._tree_update_from_subtree(self._trainables, {collection: updates})
        return self.replace(_trainables=updates)

    def set_bijector(self, collection: str, **kwargs):
        """
        Replace the bijector of parameters in a specified `collection`.

        Args:
            collection: the name of the collection that holds the target variable.
            kwargs: The name and the new bijector of the target variables within the `collection`.

        Raises:
            ValueError: if the specified `collection` is not present in the `param` `VariableDict`.

        Returns:
            A new `Param` with the updated bijectors of the variables.
        """
        if collection not in self._trainables:
            raise ValueError(f"there is no {collection} collection")

        # first update the subtree within the specified collection
        updates = self._tree_update_from_subtree(self._bijectors[collection], kwargs)
        # then update the bijectors with the newly updated collection
        updates = self._tree_update_from_subtree(self._bijectors, {collection: updates})
        return self.replace(_bijectors=updates)

    def unconstrained(self) -> "Param":
        """
        Move the `params` in the unconstrained (optimisation) space to optimise over them.

        NOTE: There is the logic to check if it is already unconstrained and return the same object,
        I just need to test it.

        Returns:
            The `Param` with the variables at the unconstrained space (the optimisation space).
        """
        # if self._constrained:
        unconstrained_params = tree_map(lambda p, t: t.inverse(p), self.params, self._bijectors)
        return self.replace(_constrained=False, params=unconstrained_params)
        # else:
        #     return self

    def constrained(self) -> "Param":
        """
        Move the `params` in the (original) constrained space.

        NOTE: There is the logic to check if it is already unconstrained and return the same object,
        I just need to test it.

        Returns:
            The `Param` with the variables at the constrained space (the original space).
        """
        # if not self._constrained:
        constrained_params = tree_map(lambda p, t: t.forward(p), self.params, self._bijectors)
        return self.replace(_constrained=True, params=constrained_params)
        # else:
        #     return self



# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is a copy from the excellent flax.training.train_state.py.

from typing import Callable, Dict

import optax

from gpfy.utils import PyTreeNode, field


class TrainState(PyTreeNode):
    """
    Simple train state for the common case with a single Optax optimizer.

    Synopsis::

        state = TrainState.create(
            apply_fn=model.apply,
            params=variables['params'],
            tx=tx)
        grad_fn = jax.grad(make_loss_fn(state.apply_fn))
        for batch in data:
          grads = grad_fn(state.params, batch)
          state = state.apply_gradients(grads=grads)

    Note that you can easily extend this dataclass by subclassing it for storing
    additional data (e.g. additional variable collections).

    For more exotic usecases (e.g. multiple optimizers) it's probably best to
    fork the class and modify it.

    Args:
      step: Counter starts at 0 and is incremented by every call to
        `.apply_gradients()`.
      apply_fn: Usually set to `model.apply()`. Kept in this dataclass for
        convenience to have a shorter params list for the `train_step()` function
        in your training loop.
      params: The parameters to be updated by `tx` and used by `apply_fn`.
      tx: An Optax gradient transformation.
      opt_state: The state for `tx`.
    """

    step: int
    apply_fn: Callable = field(pytree_node=False)
    params: Dict = field(pytree_node=True)
    tx: optax.GradientTransformation = field(pytree_node=False)
    opt_state: optax.OptState = field(pytree_node=True)

    def apply_gradients(self, *, grads, **kwargs):
        """
        Updates `step`, `params`, `opt_state` and `**kwargs` in return value.

        Note that internally this function calls `.tx.update()` followed by a call
        to `optax.apply_updates()` to update `params` and `opt_state`.

        Args:
          grads: Gradients that have the same pytree structure as `.params`.
          **kwargs: Additional dataclass attributes that should be `.replace()`-ed.

        Returns:
          An updated instance of `self` with `step` incremented by one, `params`
          and `opt_state` updated by applying `grads`, and additional attributes
          replaced as specified by `kwargs`.
        """
        updates, new_opt_state = self.tx.update(grads, self.opt_state, self.params)
        new_params = optax.apply_updates(self.params, updates)

        return self.replace(
            step=self.step + 1,
            params=new_params,
            opt_state=new_opt_state,
            **kwargs,
        )

    @classmethod
    def create(cls, *, apply_fn, params, tx, **kwargs):
        """Creates a new instance with `step=0` and initialized `opt_state`."""
        opt_state = tx.init(params)
        return cls(
            step=0,
            apply_fn=apply_fn,
            params=params,
            tx=tx,
            opt_state=opt_state,
            **kwargs,
        )



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from jax.tree_util import tree_leaves, tree_map
from jaxtyping import Array, Float

from gpfy.param import Param, identity
from gpfy.typing import BijectorDict, ConstantDict, TrainableDict, VariableDict
from gpfy.utils import dataclass, field


@dataclass
class VariationalDistribution:
    """
    Base dataclass for a variational distribution.

    Currently only Gaussian distributions are supported. Each derived class should implement the
    following abstract methods:
        * project_mean(param, ...)
        * project_diag_variance(param, ...)
        * project_variance(param, ...)
        * logdet(param, ...)
        * trace(param, ...)
        * _define_covariance_bijector()

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="VariationalDistribution", pytree_node=False)

    def __init_subclass__(cls):
        """Make sure inherting classes are dataclasses."""
        dataclass(cls)  # pytype: disable=wrong-arg-types

    def _define_covariance_bijector(self) -> tfp.bijectors.Bijector:
        """Defines the bijector to transform the covariance parameter."""
        raise NotImplementedError()

    def init(
        self,
        num_inducing_features: int,
        num_independent_processes: int = 1,
        param: Optional[Param] = None,
    ) -> Param:
        """
        Initialise the parameters of the variational distribution and return a `Param` object.

        Args:
            num_inducing_features: The number of inducing features to learn.
            num_independent_processes: The number of independent processes we need to learn.
                Used for independently modelling multiple outputs. Defaults to 1.
            param: An already initialised collection of parameters from another object, i.e., a
                kernel or spherical harmonics. If it is provided we extend the `Param` with the
                additional collection. Otherwise we return a new `Param`. Defaults to None.

        Raises:
            ValueError: Raises if the number of inducing features doesn't match the total number of
                harmonics in case we have already initialised a `Param` with `SphericalHarmonics`.

        Returns:
            The `Param` object with all variables.
        """
        inducing_features = tree_leaves(param.params.get("variational")) if param else None
        if inducing_features:
            if num_inducing_features != sum(tree_map(lambda x: x.shape[0], inducing_features)):
                raise ValueError("`param` object contains different number of inducing features.")

        # initialise the collection
        collection = "variational"
        mu = jnp.zeros((num_inducing_features, num_independent_processes), dtype=jnp.float64)
        cov = jax.vmap(lambda _: jnp.eye(num_inducing_features, dtype=jnp.float64))(
            jnp.arange(num_independent_processes)
        )
        var_params = {"mu": mu, "Sigma": cov}

        # Define the bijectors
        var_bijectors = {
            "mu": identity(),
            "Sigma": self._define_covariance_bijector(),
        }

        all_params: VariableDict = {collection: {}}
        all_trainables: TrainableDict = {collection: {}}
        all_bijectors: BijectorDict = {collection: {}}
        all_constants: ConstantDict = {}
        constrained = True

        # get the collections in case we have a provided Param
        if param:
            all_params = param.params
            all_bijectors = param._bijectors
            all_trainables = param._trainables
            all_constants = param.constants
            constrained = param._constrained

        # extend the collection and return a new param
        all_params[collection][self.name] = var_params
        all_bijectors[collection][self.name] = var_bijectors
        return Param(
            params=all_params,
            _trainables=all_trainables,
            _bijectors=all_bijectors,
            constants=all_constants,
            _constrained=constrained,
        )

    def mu(self, param: Param) -> Float[Array, "M L"]:
        """
        Helper function to easily access the mean of the distribution from the provided `Param`.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The mean of the distribution.
        """
        return param.params["variational"][self.name]["mu"]

    def cov_part(self, param: Param) -> Float[Array, "L M M"]:
        """
        Helper function to easily access the covariance (part) of the distribution from the `Param`.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The covariance parameterisation of the distribution.
        """
        return param.params["variational"][self.name]["Sigma"]

    def project_mean(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the mean of the distribution given a projection matrix.

        In standard sparse GP inference the mean of the posterior process is given by::
            m(x) = KₙₘKₘₘ⁻¹ μ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                mean of the distribution.

        Returns:
            The projected mean of the distribution.
        """
        raise NotImplementedError()

    def project_diag_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the diagonal variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The diagonal of projected covariance of the distribution.
        """
        raise NotImplementedError()

    def project_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "L N N"]:
        """
        Project the variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The projected covariance of the distribution.
        """
        raise NotImplementedError()

    def logdet(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the log determinant of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The log determinant of the covariance of the distribution.
        """
        raise NotImplementedError()

    def trace(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the trace of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The trace of the covariance of the distribution.
        """
        raise NotImplementedError()

    def prior_KL(self, param: Param) -> Float[Array, ""]:
        """
        Evaluate the KL divergence `KL(q||p)` between the variational distribution and some prior.

        NOTE: We assume we are always working on the whitened prior case, so p(u) = 𝒩(0, I).

        Returns:
            The KL divergence `KL(q||p)`.
        """
        # we are always in whitened case
        KL = -0.5 * jnp.array(jnp.size(self.mu(param)), jnp.float64)  # constant term
        KL -= 0.5 * jnp.sum(self.logdet(param))  # logdet term
        KL += 0.5 * jnp.sum(jnp.square(self.mu(param)))  # mahalanobis term
        KL += 0.5 * jnp.sum(self.trace(param))  # trace term

        return KL


class VariationalDistributionTriL(VariationalDistribution):
    """
    Variational distribution with a lower triangular parameterisation for the half-covariance.

    The distribution has the form of::
        q(u) = 𝒩(μ, LLᵀ),

    where Σ = LLᵀ.

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="VariationalDistributionTriL", pytree_node=False)

    def _define_covariance_bijector(self) -> tfp.bijectors.Bijector:
        """Return the lower triangular bijector."""
        return tfp.bijectors.FillTriangular()

    def logdet(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the log determinant of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The log determinant of the covariance of the distribution.
        """
        L = self.cov_part(param)
        return jnp.sum(jnp.log(jnp.square(jax.vmap(jnp.diag)(L))), axis=-1)

    def trace(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the trace of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The trace of the covariance of the distribution.
        """
        L = self.cov_part(param)
        return jnp.sum(jnp.square(L), axis=[-1, -2])

    def project_mean(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the mean of the distribution given a projection matrix.

        In standard sparse GP inference the mean of the posterior process is given by::
            m(x) = KₙₘKₘₘ⁻¹ μ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                mean of the distribution.

        Returns:
            The projected mean of the distribution.
        """
        mu = self.mu(param)
        return jnp.matmul(A.swapaxes(-1, -2), mu)

    def project_diag_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the diagonal variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The diagonal of projected covariance of the distribution.
        """
        L = self.cov_part(param)

        A = A[..., None, :, :]  # match any leading dims
        tmp = jnp.matmul(L.swapaxes(-1, -2), A)  # [L, M, N]
        projected_diag_variance = jnp.sum(jnp.square(tmp), -2)  # [L, N]
        return projected_diag_variance.swapaxes(-1, -2)

    def project_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "L N N"]:
        """
        Project the variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The projected covariance of the distribution.
        """
        L = self.cov_part(param)

        A = A[..., None, :, :]  # match any leading dims
        tmp = jnp.matmul(L.swapaxes(-1, -2), A)
        projected_variance = jnp.matmul(tmp.swapaxes(-1, -2), tmp)
        return projected_variance


class VariationalDistributionFullCovariance(VariationalDistribution):
    """
    Variational distribution with a full covariance parameterisation.

    The distribution has the form of::
        q(u) = 𝒩(μ, Σ).

    NOTE: There is no guarantee that the covariance will remain positive-definite if we use it in an
    optimisation loop.

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="VariationalDistributionFullCovariance", pytree_node=False)

    def _define_covariance_bijector(self) -> tfp.bijectors.Bijector:
        """Return the identity bijector."""
        return identity()

    def logdet(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the log determinant of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The log determinant of the covariance of the distribution.
        """
        S = self.cov_part(param)
        L = jnp.linalg.cholesky(S)
        return jnp.sum(jnp.log(jnp.square(jax.vmap(jnp.diag)(L))), axis=-1)

    def trace(self, param: Param) -> Float[Array, " L"]:
        """
        Evaluate the trace of the covariance of the distribution.

        Args:
            param: A `Param` initialised with the variational distribution.

        Returns:
            The trace of the covariance of the distribution.
        """
        S = self.cov_part(param)
        return jnp.trace(S, axis1=-2, axis2=-1)

    def project_mean(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the mean of the distribution given a projection matrix.

        In standard sparse GP inference the mean of the posterior process is given by::
            m(x) = KₙₘKₘₘ⁻¹ μ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                mean of the distribution.

        Returns:
            The projected mean of the distribution.
        """
        mu = self.mu(param)
        return jnp.matmul(A.swapaxes(-1, -2), mu)

    def project_diag_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "N L"]:
        """
        Project the diagonal variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The diagonal of projected covariance of the distribution.
        """
        S = self.cov_part(param)

        A = A[..., None, :, :]  # match any leading dims
        tmp = jnp.matmul(S, A)
        projected_diag_variance = jnp.sum(tmp * A, -2)
        return projected_diag_variance.swapaxes(-1, -2)

    def project_variance(self, param: Param, A: Float[Array, "M N"]) -> Float[Array, "L N N"]:
        """
        Project the variance of the distribution given a projection matrix.

        In standard sparse GP inference the covariance of the posterior process is given by::
            var(x) = Kₙₙ - KₙₘKₘₘ⁻¹Kₘₙ,
        where Kₘₘ⁻¹Kₘₙ is regarded to be the projection.

        Args:
            param: A `Param` initialised with the variational distribution.
            A: A projection, normally coming from the model/process side, that is applied to the
                covariance of the distribution.

        Returns:
            The projected covariance of the distribution.
        """
        S = self.cov_part(param)

        A = A[..., None, :, :]  # match any leading dims
        tmp = jnp.matmul(S, A)
        projected_variance = jnp.matmul(A.swapaxes(-1, -2), tmp)
        return projected_variance



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import jax.numpy as jnp
from jaxtyping import Array, Float
from scipy.special import loggamma, roots_legendre

from gpfy.gegenbauer import gegenbauer

legendre_x, legendre_w = roots_legendre(1000)
legendre_x = jnp.array(legendre_x, dtype=jnp.float64)
legendre_w = jnp.array(legendre_w, dtype=jnp.float64)


def weight_func(t: Float[Array, " N"], d: int) -> Float[Array, " N"]:
    """
    The function which orthogonalises Gegenbauer polynomials in [-1, 1].

    Args:
        t: The input to the weight function and the Gegenbauer polynomial.
        d: The dimensionality of the input on the Sᵈ⁻¹ sphere.

    Returns:
        The function evaluation at `t`.
    """
    return (1.0 - jnp.square(t)) ** ((d - 3) / 2)


def funk_hecke_lambda(
    fn: Callable[[Float[Array, " N"]], Float[Array, " N"]], n: int, d: int
) -> Float[Array, ""]:
    """
    The general Funk-Hecke formula that gives the `n`th eigenvalue of function `fn` on Sᵈ⁻¹.

    It computes the integral::
        λₙ = ω / Cᵅₙ(1) ∫₋₁¹ f(t) Cᵅₙ(t) (1 - t²)ʳ dt,
    where::
        α = ⁽ᵈ⁻²⁾⁄₂, r = α - ¹⁄₂
    and::
        ω = |Sᵈ⁻²| / |Sᵈ⁻¹|

    Args:
        fn: The function which we integrate.
        n: The order of the eigenvalue
        d: The dimensionality of the input on the Sᵈ⁻¹ sphere.

    Returns:
        The nth eigenvalue.
    """
    alpha = (d - 2.0) / 2.0
    C1 = gegenbauer(n, alpha, jnp.array(1, dtype=jnp.float64))
    solid_angle_ratio = jnp.exp(loggamma(d / 2) - loggamma((d - 1) / 2)) / jnp.sqrt(jnp.pi)

    def integrand(x):
        return gegenbauer(n, alpha, x) * fn(x) * weight_func(x, d)

    integral = jnp.dot(legendre_w, integrand(legendre_x))
    return solid_angle_ratio / C1 * integral



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from gpfy.param import Param, identity, positive
from gpfy.spherical import NTK, ArcCosine, PolynomialDecay

key = jr.PRNGKey(42)

kernels = [ArcCosine, NTK, PolynomialDecay]
dims = [2, 3, 5, 10]
order_list = [0, 1, 2]
ards = [True, False]
depths = [1, 2, 5]
projection_dims = [None, 2]
shape_list = [(1, 3), (10, 7)]


def _get_input_from_shape(shape):
    return jr.normal(key, shape, dtype=jnp.float64)


def _create_kernel(kernel, order, ard, depth):
    if kernel.__name__ == "ArcCosine":
        return kernel(order=order, ard=ard, depth=depth)
    elif kernel.__name__ == "NTK":
        return kernel(ard=ard, depth=depth)
    else:
        return kernel(ard=ard)


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("order", order_list)
@pytest.mark.parametrize("ard", ards)
@pytest.mark.parametrize("depth", depths)
@pytest.mark.parametrize("input_dim", dims)
@pytest.mark.parametrize("projection_dim", projection_dims)
def test_kernel_init(kernel, order, ard, depth, input_dim, projection_dim):
    """Test the kernel initialisation."""
    if kernel.__name__ != "ArcCosine" and order != 1:
        return
    if kernel.__name__ == "PolynomialDecay" and depth != 1:
        return
    k = _create_kernel(kernel, order, ard, depth)
    param = k.init(key, input_dim=input_dim, projection_dim=projection_dim)
    collection = k.name  # the kernel name

    assert collection in param.params
    assert collection in param._bijectors
    assert collection in param._trainables
    assert "sphere" in param.constants

    variables = ["weight_variances", "bias_variance", "variance"]
    for v in variables:
        assert v in param.params[collection].keys()
        gt_bij = identity if v == "weight_variances" and projection_dim else positive
        assert isinstance(param._bijectors[collection][v], gt_bij)

    if collection == "PolynomialDecay":
        assert "beta" in param.params[collection].keys()
        assert isinstance(param._bijectors[collection]["beta"], positive)

    if projection_dim:
        assert param.params[collection]["weight_variances"].shape == (input_dim, projection_dim)
    else:
        if ard:
            assert param.params[collection]["weight_variances"].shape == (input_dim,)
        else:
            assert param.params[collection]["weight_variances"].shape == ()

    if kernel.__name__ == "PolynomialDecay":
        "gegenbauer_lookup_table" in param.constants["sphere"]
    else:
        "eigenvalues" in param.constants["sphere"]


@pytest.mark.parametrize("kernel", kernels)
def test_raises_for_inappropriate_order(kernel):
    """Test we error when we pass incompatible order value given the kernel class."""
    if kernel.__name__ == "ArcCosine":
        with pytest.raises(ValueError):
            kernel(order=4)
    else:
        with pytest.raises(TypeError):
            kernel(order=2)


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("shape", shape_list)
@pytest.mark.parametrize("projection_dim", projection_dims)
def test_to_sphere_returns_unit_length_vector(kernel, shape, projection_dim):
    """Test the unit length of the projected vectors."""
    x = _get_input_from_shape(shape)
    dim = x.shape[-1]
    k = _create_kernel(kernel, 1, True, 3)

    param = k.init(key, input_dim=dim, projection_dim=projection_dim)

    x_sphere, radius = k.to_sphere(param, x)

    assert x_sphere.shape[-1] == param.constants["sphere"]["sphere_dim"] + 1
    np.testing.assert_allclose(jnp.sum(x_sphere**2, -1), jnp.ones_like(radius))

    x_scaled = k._scale_X(param, x)
    np.testing.assert_allclose(jnp.sum(x_scaled**2, -1), radius**2 - 1)  # bias initialised to 1


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("shape", shape_list)
@pytest.mark.parametrize("projection_dim", projection_dims)
def test_scaled_X_returns_expected_shape(kernel, shape, projection_dim):
    """Test the shape of the projected vectors."""
    x = _get_input_from_shape(shape)
    dim = x.shape[-1]
    k = _create_kernel(kernel, 1, True, 3)

    param = k.init(key, input_dim=dim, projection_dim=projection_dim)
    x_scaled = k._scale_X(param, x)
    assert x_scaled.shape[-1] == (projection_dim or dim)


@pytest.mark.parametrize("kernel", kernels[:-1])
def test_eigenvalues_return_zero_if_not_precomputed(kernel):
    """Test we return 0 for non-precomputed eigvals."""
    k = _create_kernel(kernel, 1, True, 3)
    dummy_param = Param()
    np.testing.assert_allclose(k.eigenvalues(dummy_param, jnp.arange(10)), 0.0)


def test_eigenvalues_are_normalised():
    """Test the eigvals are normalised for the PolynomialDecay kernel."""
    k = _create_kernel(PolynomialDecay, 1, True, 3)
    param = k.init(key, input_dim=5)

    sphere_dim = param.constants["sphere"]["sphere_dim"]
    geg = param.constants["sphere"]["gegenbauer_lookup_table"]
    const_factor = k._compute_eigvals(sphere_dim, gegenbauer_lookup_table=geg)

    np.testing.assert_allclose(jnp.sum(k.eigenvalues(param, jnp.arange(10)) / const_factor), 1.0)


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("order", order_list)
@pytest.mark.parametrize("depth", depths)
def test_shape_function_is_normalised(kernel, order, depth):
    """Test the shape_function is normalised so that κ(1) = 1."""
    k = _create_kernel(kernel, order, True, depth)
    param = k.init(key, input_dim=3)

    np.testing.assert_allclose(k.shape_function(param, jnp.ones((1, 1))), 1.0)


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("order", order_list)
@pytest.mark.parametrize("ard", ards)
@pytest.mark.parametrize("depth", depths)
@pytest.mark.parametrize("projection_dim", projection_dims)
def test_K_vs_K_diag(kernel, order, ard, depth, projection_dim):
    """Test that k.K(X, X) and k.K_diag(X) give the same answer for the diagonal entries."""
    if kernel.__name__ != "ArcCosine" and order != 1:
        return
    if kernel.__name__ == "PolynomialDecay" and depth != 1:
        return

    x = _get_input_from_shape((10, 5))
    k = _create_kernel(kernel, order, ard, depth)
    param = k.init(key, input_dim=x.shape[-1], projection_dim=projection_dim)

    np.testing.assert_allclose(jnp.diag(k.K(param, x)), k.K_diag(param, x))


@pytest.mark.parametrize("kernel", kernels)
@pytest.mark.parametrize("order", [1, 2])
@pytest.mark.parametrize("ard", ards)
@pytest.mark.parametrize("depth", depths)
@pytest.mark.parametrize("projection_dim", projection_dims)
def test_K(kernel, order, ard, depth, projection_dim):
    """Test that k.K(X) and k.K(X, X) give the same answer."""
    if kernel.__name__ != "ArcCosine" and order != 1:
        return
    if kernel.__name__ == "PolynomialDecay" and depth != 1:
        return

    x = _get_input_from_shape((10, 5))
    k = _create_kernel(kernel, order, ard, depth)
    param = k.init(key, input_dim=x.shape[-1], projection_dim=projection_dim)

    np.testing.assert_allclose(k.K(param, x), k.K(param, x, x))
    np.testing.assert_allclose(k.K(param, x), k.K2(param, x))
    np.testing.assert_allclose(k.K(param, x), k.K2(param, x, x))



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from gpfy.likelihoods import Gaussian
from gpfy.model import GP
from gpfy.spherical import NTK, PolynomialDecay
from gpfy.spherical_harmonics import SphericalHarmonics
from gpfy.variational import VariationalDistributionTriL

key = jr.PRNGKey(42)

dims = [3, 5]
num_independent_processes = [1, 7]


def _create_model_and_variational():
    k = PolynomialDecay()
    k = NTK()
    m = GP(kernel=k)
    sh = SphericalHarmonics(5, 10)
    q = VariationalDistributionTriL()
    return m, sh, q


def test_conditional_model_has_non_empty_tuple():
    m, sh, q = _create_model_and_variational()
    assert not m.conditional_fn  # assert tuple is empty

    m = m.conditional(sh, q)
    assert m.conditional_fn


@pytest.mark.parametrize("num_processes", num_independent_processes)
@pytest.mark.parametrize("input_dim", dims)
def test_predict_mean(input_dim, num_processes):
    m, sh, q = _create_model_and_variational()
    lik = Gaussian()
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    x = jr.normal(key, (10, input_dim), dtype=jnp.float64)
    # predict from prior should be 0
    mu = m.mu(param)(x)
    np.testing.assert_allclose(mu, 0)
    np.testing.assert_array_equal(mu.shape, (10, 1))

    # condition on the var features
    m = m.conditional(sh, q)
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    mu = m.mu(param)(x)
    np.testing.assert_array_equal(mu.shape, (10, num_processes))


@pytest.mark.parametrize("num_processes", num_independent_processes)
@pytest.mark.parametrize("input_dim", dims)
def test_predict_var(input_dim, num_processes):
    m, sh, q = _create_model_and_variational()
    lik = Gaussian()
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    x = jr.normal(key, (10, input_dim), dtype=jnp.float64)
    # predict from prior should be k.K_diag
    var = m.var(param)(x)
    np.testing.assert_allclose(var, m.kernel.K_diag(param, x))
    np.testing.assert_array_equal(var.shape, (10,))

    # condition on the var features
    m = m.conditional(sh, q)
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    var = m.var(param)(x)
    np.testing.assert_array_equal(var.shape, (10, num_processes))


@pytest.mark.parametrize("num_processes", num_independent_processes)
@pytest.mark.parametrize("input_dim", dims)
def test_predict_cov(input_dim, num_processes):
    m, sh, q = _create_model_and_variational()
    lik = Gaussian()
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    x = jr.normal(key, (10, input_dim), dtype=jnp.float64)
    # predict from prior should be k.K
    cov = m.cov(param)(x)
    np.testing.assert_allclose(cov, m.kernel.K(param, x))
    np.testing.assert_array_equal(cov.shape, (10, 10))

    # condition on the var features
    m = m.conditional(sh, q)
    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    cov = m.cov(param)(x)
    np.testing.assert_array_equal(cov.shape, (num_processes, 10, 10))


@pytest.mark.parametrize("num_processes", num_independent_processes)
@pytest.mark.parametrize("input_dim", dims)
def test_predict_diag_same_as_mu_and_var(input_dim, num_processes):
    m, sh, q = _create_model_and_variational()
    m = m.conditional(sh, q)

    lik = Gaussian()
    x = jr.normal(key, (10, input_dim), dtype=jnp.float64)

    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    mu = m.mu(param)(x)
    var = m.var(param)(x)
    mu2, var2 = m.predict_diag(param, x)
    np.testing.assert_allclose(mu, mu2)
    np.testing.assert_allclose(var, var2)


@pytest.mark.parametrize("num_processes", num_independent_processes)
@pytest.mark.parametrize("input_dim", dims)
def test_predict_same_as_mu_and_cov(input_dim, num_processes):
    m, sh, q = _create_model_and_variational()
    m = m.conditional(sh, q)

    lik = Gaussian()
    x = jr.normal(key, (10, input_dim), dtype=jnp.float64)

    param = m.init(
        key,
        input_dim,
        num_processes,
        num_processes,
        likelihood=lik,
        sh_features=sh,
        variational_dist=q,
    )

    mu = m.mu(param)(x)
    cov = m.cov(param)(x)
    mu2, cov2 = m.predict(param, x)
    np.testing.assert_allclose(mu, mu2)
    np.testing.assert_allclose(cov, cov2)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from gpfy.spherical_harmonics import SphericalHarmonics
from gpfy.variational import (
    VariationalDistributionFullCovariance,
    VariationalDistributionTriL,
)

key = jr.PRNGKey(42)

num_inducing_features_list = [2, 5, 10]
num_independent_processes_list = [1, 7]
var_classes = [VariationalDistributionFullCovariance, VariationalDistributionTriL]


def test_init_fails_when_passing_param_with_wrong_size_of_inducing_features():
    sh = SphericalHarmonics(5, 10)
    sh_param = sh.init(key, input_dim=4)
    num_inducing = sh.num_inducing(sh_param)
    q = VariationalDistributionTriL()
    with pytest.raises(ValueError):
        _ = q.init(num_inducing_features=num_inducing + 1, param=sh_param)


@pytest.mark.parametrize("num_inducing", num_inducing_features_list)
@pytest.mark.parametrize("num_processes", num_independent_processes_list)
@pytest.mark.parametrize("var_class", var_classes)
def test_init_create_correct_variables(num_inducing, num_processes, var_class):
    q = var_class()
    param = q.init(num_inducing_features=num_inducing, num_independent_processes=num_processes)

    assert "mu" in param.params["variational"][q.name]
    np.testing.assert_array_equal(q.mu(param).shape, (num_inducing, num_processes))
    assert "Sigma" in param.params["variational"][q.name]
    np.testing.assert_array_equal(
        q.cov_part(param).shape, (num_processes, num_inducing, num_inducing)
    )


@pytest.mark.parametrize("num_inducing", num_inducing_features_list)
@pytest.mark.parametrize("num_processes", num_independent_processes_list)
@pytest.mark.parametrize("var_class", var_classes)
def test_log_det(num_inducing, num_processes, var_class):
    q = var_class()
    param = q.init(num_inducing_features=num_inducing, num_independent_processes=num_processes)
    num_processes = q.mu(param).shape[-1]

    # initialised to identity so logdet should be 0.
    np.testing.assert_array_equal(q.logdet(param), jnp.zeros((num_processes,)))


@pytest.mark.parametrize("num_inducing", num_inducing_features_list)
@pytest.mark.parametrize("num_processes", num_independent_processes_list)
@pytest.mark.parametrize("var_class", var_classes)
def test_trace(num_inducing, num_processes, var_class):
    q = var_class()
    param = q.init(num_inducing_features=num_inducing, num_independent_processes=num_processes)
    num_inducing, num_processes = q.mu(param).shape

    # initialised to identity so trace should be num_inducing.
    np.testing.assert_array_equal(q.trace(param), num_inducing * jnp.ones((num_processes,)))


@pytest.mark.parametrize("num_inducing", num_inducing_features_list)
@pytest.mark.parametrize("num_processes", num_independent_processes_list)
@pytest.mark.parametrize("var_class", var_classes)
def test_projection_shapes(num_inducing, num_processes, var_class):
    q = var_class()
    param = q.init(num_inducing_features=num_inducing, num_independent_processes=num_processes)
    num_inducing, num_processes = q.mu(param).shape

    x = jr.normal(key, (num_inducing, 13))

    proj = q.project_mean(param, x)
    np.testing.assert_array_equal(proj.shape, (x.shape[-1], num_processes))

    proj = q.project_variance(param, x)
    np.testing.assert_array_equal(proj.shape, (num_processes, x.shape[-1], x.shape[-1]))

    proj = q.project_diag_variance(param, x)
    np.testing.assert_array_equal(proj.shape, (x.shape[-1], num_processes))



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from jaxtyping import Array, Float

from gpfy.param import Param, positive
from gpfy.quadrature import gauss_hermite_quadrature
from gpfy.typing import BijectorDict, ConstantDict, TrainableDict, VariableDict
from gpfy.utils import dataclass, field


@dataclass
class Likelihood:
    """
    The base class for a Likelihood object.

    We rely on tensorflow_probability for the heavy work.

    Currently only a Gaussian and a Bernoulli distribution are supported.
    Each derived class should implement the following abstract methods:
        * conditional_distribution(param, ...)
        * variational_expectations(param, ...)

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="Likelihood", pytree_node=False)

    def __init_subclass__(cls):
        """Make sure inherting classes are dataclasses."""
        dataclass(cls)  # pytype: disable=wrong-arg-types

    def init(self, param: Optional[Param] = None) -> Param:
        """
        Iinitialise the parameters, if any, of the likelihood.

        Args:
            param: An already initialised collection of parameters from another object, i.e., a
                kernel or spherical harmonics. If it is provided we extend the `Param` with the
                additional collection. Otherwise we return a new `Param`. Defaults to None.

        Returns:
            The `Param` object with all variables.
        """
        # basic set-up
        return param or Param()

    def log_prob(
        self, param: Param, F: Float[Array, "N D"], Y: Float[Array, "N D"]
    ) -> Float[Array, " N"]:
        """
        The log probability density log p(Y|F), where we sum over the independent output dimensions.

        Args:
            param: A `Param` initialised with the likelihood.
            F: The ranodm vector, which is usually a sample obtained from the GP.
            Y: The outputs we want to match via the distribution.

        Returns:
            The log probability density.
        """
        return jnp.sum(self.conditional_distribution(param, F).log_prob(Y), -1)

    def conditional_distribution(
        self, param: Param, F: Float[Array, "N D"]
    ) -> tfp.distributions.Distribution:
        """
        The distribution of the outputs conditioned on a latent function evaluation, i.e., p(Y|F).

        Args:
            param: A `Param` initialised with the likelihood.
            F: The ranodm vector, which is usually a sample obtained from the GP.

        Returns:
            The conditioanl distribution.
        """
        raise NotImplementedError()

    def variational_expectations(
        self, param: Param, Fmu: Float[Array, "N D"], Fvar: Float[Array, "N D"]
    ) -> Callable[[Float[Array, "N D"]], Float[Array, " N"]]:
        """
        Define the expected log density, given a Gaussian distribution for the function values.

        In variational inference we have a variational posterior over functions
            q(f) = 𝒩(Fmu, Fvar).

        Then for a liklihood

            p(y|f)

        we compute

           ∫ log(p(y=Y|f)) q(f) df.

        Returns:
            The expected log density function that can be applied on some data `Y`.
        """
        raise NotImplementedError()


class Gaussian(Likelihood):
    """
    The Gaussian likelihood parameterised by the noise variance.

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="Gaussian", pytree_node=False)

    def init(self, param: Optional[Param] = None) -> Param:
        """
        Initialise the parameters of the Gaussian likelihood and return a `Param` object.

        Args:
            param: An already initialised collection of parameters from another object, i.e., a
                kernel or spherical harmonics. If it is provided we extend the `Param` with the
                additional collection. Otherwise we return a new `Param`. Defaults to None.

        Returns:
            The `Param` object with all variables.
        """
        lik_params = {"variance": jnp.array(1.0, dtype=jnp.float64)}
        lik_bijectors = {"variance": positive()}

        # initialise the collection
        collection = "likelihood"
        all_params: VariableDict = {}
        all_trainables: TrainableDict = {collection: {}}
        all_bijectors: BijectorDict = {}
        all_constants: ConstantDict = {}
        constrained = True

        # get the collections in case we have a provided Param
        if param:
            all_params = param.params
            all_bijectors = param._bijectors
            all_trainables = param._trainables
            all_constants = param.constants
            constrained = param._constrained

        # extend the collection and return a new param
        all_params[collection], all_bijectors[collection] = {}, {}
        all_params[collection][self.name] = lik_params
        all_bijectors[collection][self.name] = lik_bijectors
        return Param(
            params=all_params,
            _trainables=all_trainables,
            _bijectors=all_bijectors,
            constants=all_constants,
            _constrained=constrained,
        )

    def conditional_distribution(
        self, param: Param, F: Float[Array, "N D"]
    ) -> tfp.distributions.Distribution:
        """
        The Normal distribution p(Y|F) = 𝒩(F, σ²).

        Args:
            param: A `Param` initialised with the likelihood.
            F: The ranodm vector, which is usually a sample obtained from the GP.

        Returns:
            The conditioanl distribution.
        """
        variance = param.params["likelihood"][self.name]["variance"]
        return tfp.distributions.Normal(loc=F.astype(jnp.float64), scale=jnp.sqrt(variance))

    def variational_expectations(
        self, param: Param, Fmu: Float[Array, "N D"], Fvar: Float[Array, "N D"]
    ) -> Callable[[Float[Array, "N D"]], Float[Array, " N"]]:
        """
        Define the expected log density, given a Gaussian distribution for the function values.

        In variational inference we have a variational posterior over functions
            q(f) = 𝒩(Fmu, Fvar).

        Then for a liklihood

            p(y|f)

        we compute

           ∫ log(p(y=Y|f)) q(f) df.

        NOTE: For the Guassian likelihood the variational expectations can be computed analytically.

        Returns:
            The expected log density function that can be applied on some data `Y`.
        """
        variance = param.params["likelihood"][self.name]["variance"]

        def _var_exp(Y: Float[Array, "N D"]):
            return jnp.sum(
                -0.5 * jnp.log(2 * jnp.pi)
                - 0.5 * jnp.log(variance)
                - 0.5 * ((Y - Fmu) ** 2 + Fvar) / variance,
                axis=-1,
            )

        return _var_exp


class Bernoulli(Likelihood):
    """
    The Bernoulli likelihood for classification.

    It is parameterised by a latent function (the GP), which is mapped via the inverse probit to the
    probability space [0, 1].

    Attributes:
        name: The name for the object. Defautls to the class name.
    """

    name: str = field(default="Bernoulli", pytree_node=False)

    def conditional_distribution(
        self, param: Param, F: Float[Array, "N D"]
    ) -> tfp.distributions.Distribution:
        """
        The distribution of the outputs conditioned on a latent function evaluation, i.e., p(Y|F).

        NOTE: The inverse probit is used as a link function to map the continuous latent function
        to the probability space [0, 1].

        Args:
            param: A `Param` initialised with the likelihood.
            F: The ranodm vector, which is usually a sample obtained from the GP.

        Returns:
            The conditioanl distribution.
        """
        return tfp.distributions.Bernoulli(probs=inv_probit(F))

    def variational_expectations(
        self, param: Param, Fmu: Float[Array, "N D"], Fvar: Float[Array, "N D"]
    ) -> Callable[[Float[Array, "N D"]], Float[Array, " N"]]:
        """
        Define the expected log density, given a Gaussian distribution for the function values.

        In variational inference we have a variational posterior over functions
            q(f) = 𝒩(Fmu, Fvar).

        Then for a liklihood

            p(y|f)

        we compute

           ∫ log(p(y=Y|f)) q(f) df.

        NOTE: For the Bernoulli likelihood the variational expectations are estimated via
        Gauss-Hermite quadrature integration.

        Returns:
            The expected log density function that can be applied on some data `Y`.
        """
        log_prob = lambda f, y: self.log_prob(param, f, y)
        return gauss_hermite_quadrature(log_prob, Fmu, jnp.sqrt(Fvar))


def inv_probit(x: Float[Array, "N D"]) -> Float[Array, "N D"]:
    """
    Compute the inverse probit function.

    Args:
        x: A vector of values.

    Returns:
        The inverse probit of the input vector.
    """
    jitter = 1e-3  # To ensure output is in interval (0, 1).
    return 0.5 * (1.0 + jax.scipy.special.erf(x / jnp.sqrt(2.0))) * (1 - 2 * jitter) + jitter



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from jax import random
from jaxtyping import Array, Float, Int

from gpfy.gegenbauer import GegenbauerLookupTable
from gpfy.harmonics.utils import funk_hecke_lambda
from gpfy.param import Param
from gpfy.typing import PRNG, ActiveDims, ConstantDict
from gpfy.utils import dataclass, field


def _slice_or_list(value: Optional[ActiveDims] = None) -> Union[ActiveDims, Array]:
    """
    Retrun a slice for indexing dimensions of inputs.

    Args:
        value: The slice to the dimensions we want to index. Defaults to None.

    Returns:
        The slice.
    """
    if value is None:
        return slice(None, None, None)
    if isinstance(value, slice):
        return value
    return jnp.array(value, dtype=int)


@dataclass
class Spherical:
    """
    Abstract base class for spherical kernels that mimic the behaviour of neural networks.

    Implementations of `ArcCosine`, `NTK` and a custom `PolynomialDecay` kernel of continuous depth.

    Key reference is
    ::
        @article{cho2009kernel,
            title={Kernel Methods for Deep Learning},
            author={Cho, Youngmin and Saul, Lawrence},
            journal={Advances in Neural Information Processing Systems},
            volume={22},
            year={2009}
        }

    NOTE: All classess inherting from `Spherical` need to implement the `shape_function(...)`.

    Attributes:
        order: The order of the spherical kernel that specifies the activation function of the
            equivalent neural network. The function is a monmial of the specified order.
            Defaults to 1.
        ard: Flag to indicate if we want to model separate weights per input dimension.
            Defaults to True.
        active_dims: Optional slice to specify the active dimensions for the kernel.
            Currently, not supported.
        name: The name for the object. Defautls to `"Spherical"`.
    """

    order: int = field(default=1, pytree_node=False)
    ard: bool = field(default=True, pytree_node=False)
    active_dims: Optional[ActiveDims] = field(default_factory=_slice_or_list, pytree_node=False)
    name: str = field(default="Spherical", pytree_node=False)

    def __init_subclass__(cls):
        """Make sure inherting classes are dataclasses."""
        dataclass(cls)  # pytype: disable=wrong-arg-types

    def init(
        self,
        key: PRNG,
        input_dim: int,
        projection_dim: Optional[int] = None,
        max_num_eigvals: Optional[int] = None,
    ) -> Param:
        """
        Initialise the parameters of the kernel and return a `Param` object.

        Args:
            key: A random key for initialising the weights.
            input_dim: The input dimension. We then append an extra dimension for bias.
            projection_dim: If specified it denotes a projection of the `input_dim` to
                `projection_dim`. Defaults to None.
            max_num_eigvals: If specified it denotes the higher order of frequency up to which we
                compute the eigenvalues. Defaults to None.

        Returns:
            The `Param` object with all variables.
        """
        bias_variance = jnp.array(1.0, dtype=jnp.float64)
        variance = jnp.array(1.0, dtype=jnp.float64)
        bijectors = {}
        if not projection_dim:
            if self.ard:
                weight_variances = jnp.ones(input_dim, dtype=jnp.float64)
            else:
                weight_variances = jnp.array(1.0, dtype=jnp.float64)
        else:
            weight_variances = random.normal(key, (input_dim, projection_dim), dtype=jnp.float64)
            bijectors = {"weight_variances": tfp.bijectors.Identity()}

        params = {
            "weight_variances": weight_variances,
            "bias_variance": bias_variance,
            "variance": variance,
        }
        # set the collection of the parameters to the name of the kernel.
        collection = self.name
        # Compute constants regarding the sphere and the spectral properties of the kernel.
        # Add an extra dimension for the bias.
        sphere_dim = projection_dim or input_dim
        dim = sphere_dim + 1
        alpha = (dim - 2.0) / 2.0
        constants: ConstantDict = {}
        constants["sphere"] = {"sphere_dim": sphere_dim, "alpha": alpha}
        constants[self.name] = {}

        if isinstance(self, PolynomialDecay):
            # parameter to control the decay
            params["beta"] = jnp.array(1.0, dtype=jnp.float64)
            gegenbauer_lookup = GegenbauerLookupTable(self.truncation_level, alpha)
            constants["sphere"]["gegenbauer_lookup_table"] = gegenbauer_lookup
        else:
            # Precompute the eigenvalues of the kernel
            eigenvalues = self._compute_eigvals(sphere_dim, max_num_eigvals=max_num_eigvals)
            constants[self.name]["eigenvalues"] = eigenvalues

        # The default bijector is positive, so we only need to pass the bijector for the weights
        # in case it's Identity due to the linear projection
        return Param(
            params={collection: params},
            _bijectors={collection: bijectors},
            constants=constants,
        )

    def _compute_eigvals(
        self,
        sphere_dim: int,
        *,
        max_num_eigvals: Optional[int] = None,
        gegenbauer_lookup_table: Optional[GegenbauerLookupTable] = None,
    ) -> Float[Array, " N"]:
        """
        Compute the eigenvalues of the spherical kernel.

        Args:
            sphere_dim: The dimensionality of the sphere.
            max_num_eigvals: The maximum number of eigenvalues to precompute. Defaults to None,
                which will compute eigenvalues up to the 20th frequency.
            gegenbauer_lookup_table: A precomputed lookup table to evaluate the Gegenbauer
                polynomial. Defaults to None.

        Returns:
            Array with the eigenvalues of the kernel.
        """
        max_num_eigvals = max_num_eigvals or 20
        # The dim of the inputs
        dim = sphere_dim + 1
        # shape function expects a param input arg, but it is not required for ArcCosine and NTK.
        dummy_param = Param()
        part_shape_function = lambda x: self.shape_function(dummy_param, x=x)
        part_funk_hecke = lambda n: funk_hecke_lambda(part_shape_function, n, dim)
        return jax.vmap(part_funk_hecke)(jnp.arange(max_num_eigvals, dtype=jnp.int32))

    def eigenvalues(self, param: Param, levels: Int[Array, " N"]) -> Float[Array, " N"]:
        """
        Get the eigenvalues from the provided `param` object, or return 0s if not precomputed.

        Args:
            param: A `Param` initialised with the kernel.
            levels: An array specifying up to which order de we need eigenvalues.

        Returns:
            The eigenvalues.
        """
        zeros = jnp.zeros((len(levels),), dtype=jnp.float64)
        eigval = param.constants.get(self.name, {}).get("eigenvalues", zeros)
        return eigval[levels]

        # if self.name in param.constants and "eigenvalues" in param.constants[self.name]:
        #     # The kernel is initalised and we have already precomputed the eigenvalues
        #     eigval = param.constants[self.name]["eigenvalues"]
        #     levels = jnp.clip(levels, 0, len(eigval) - 1)
        # else:
        #     # we have a PolynomailDecay kernel and we need to compute the eigenvalues on the spot.
        #     assert isinstance(self, PolynomialDecay)  # helps mypy
        #     n = jnp.arange(self.truncation_level, dtype=jnp.float64)
        #     beta = param.params[self.name]["beta"]
        #     sphere_dim = param.constants["sphere"]["sphere_dim"]
        #     geg = param.constants["sphere"]["gegenbauer_lookup_table"]
        #     decay = (1 + n) ** (-beta)
        #     const_factor = self._compute_eigvals(sphere_dim, gegenbauer_lookup_table=geg)
        #     eigval = decay * const_factor / jnp.sum(decay)
        #     levels = jnp.clip(levels, 0, self.truncation_level - 1)
        # return eigval[levels]

    def _scale_X(self, param: Param, X: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        Scale the input X by the weights.

        Args:
            param: The `Param` initialised with the kernel.
            X: Input Array already projected to the unit sphere.

        Returns:
            The scaled input.
        """
        weight_variances = param.params[self.name]["weight_variances"]
        if len(weight_variances.shape) == 2:
            return jnp.matmul(X, weight_variances)
        return X * jnp.sqrt(weight_variances)

    def to_sphere(
        self, param: Param, X: Float[Array, "N D"]
    ) -> Tuple[Float[Array, "N DSphere"], Float[Array, " N"]]:
        """
        Normalise the input X and project it on the unit sphere.

        Args:
            param: The `Param` initialised with the kernel.
            X: Input Array.

        Returns:
            A tuple containing the normalised input and the norm (radius) of the vector.
        """
        scaled_X = self._scale_X(param, X)
        bias_shape = scaled_X.shape[:-1] + (1,)
        b = param.params[self.name]["bias_variance"]
        bias = jnp.ones(bias_shape, dtype=jnp.float64) * jnp.sqrt(b)
        X_with_bias = jnp.concatenate([scaled_X, bias], axis=-1)
        r = jnp.sqrt(jnp.sum(jnp.square(X_with_bias), axis=-1))
        return X_with_bias / r[..., None], r

    def shape_function(self, param: Param, x: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        The shape function that determines the equivalent kernel to a deep network architecture.

        Args:
            param: The `Param` initialised with the kernel.
            x: Input Array.

        Raises:
            NotImplementedError: All derived classes need to implement this method.
        """
        raise NotImplementedError

    def _squash(self, x: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        Ensure that the normalised input is squashed between `[-1, 1]`.

        Due to numerical precision the input can be larger than 1, which causes NaNs when evaluating
        the Gegenbauer polynomial.

        Args:
            x: Input Array.

        Returns:
            The squashed input.
        """
        return jnp.clip(x, -1.0, 1.0)

    def kappa(self, u: Float[Array, " N"], order: int) -> Float[Array, " N"]:
        """
        Compute the analytic integral of the angular part for angles with cosine `u`.

        This is essentially the kernel evaluation.

        NOTE: We have only implemented the first 3 orders which correspond to:
            0 -> step function
            1 -> ReLU
            2 -> Similar to elu

        NOTE: It ignores any specified order higher than 2 and clips it back to 2.

        Args:
            u: The cosine of the angle between normalised vectors on the unit sphere.
            order: The order of the kernel which determines the equivalent activation function.

        Returns:
            The shape function evaluated at the specified angles.
        """

        def _zero_order():
            return (jnp.pi - jnp.arccos(u)) / jnp.pi

        def _first_order():
            return (u * (jnp.pi - jnp.arccos(u)) + jnp.sqrt(1.0 - jnp.square(u))) / jnp.pi

        def _second_order():
            return (
                (1.0 + 2.0 * jnp.square(u)) / 3 * (jnp.pi - jnp.arccos(u))
                + jnp.sqrt(1.0 - jnp.square(u))
            ) / jnp.pi

        return jax.lax.switch(order, (_zero_order, _first_order, _second_order))

    def K(
        self,
        param: Param,
        X: Float[Array, "N D"],
        X2: Optional[Float[Array, "M D"]] = None,
    ) -> Float[Array, "N M"]:
        """
        Evaluates the covariances element-wise between the all pairs in `X` and `X2`.

        Similar to `self.K2` but with broadcasting instead of vmap.
        NOTE: If `X2` is not specified, the method evaluates the covariances between pairs of `X`.

        Args:
            param: The `Param` initialised with the kernel.
            X: Input Array.
            X2: Input Array.

        Returns:
            The covariance between the specified intpus.
        """
        # project inputs
        X_sphere, rad1 = self.to_sphere(param, X)
        if X2 is None:  # we compute the covariance between pairs of X, so copy X2 <- X
            X_sphere2 = X_sphere
            rad2 = rad1
            K = jax.vmap(lambda x: jax.vmap(lambda y: jnp.dot(x, y))(X_sphere2))(X_sphere)
            i, j = jnp.diag_indices(K.shape[-1])
            K = K.at[..., i, j].set(1.0)  # make sure diagonal is 1.
        else:  # proceed as normal and compute covariance between X, X2
            # project X2 on the sphere
            X_sphere2, rad2 = self.to_sphere(param, X2)
            K = jax.vmap(lambda x: jax.vmap(lambda y: jnp.dot(x, y))(X_sphere2))(X_sphere)

        K = self.shape_function(param, K)
        r = jax.vmap(lambda x: jax.vmap(lambda y: jnp.dot(x, y))(rad2))(rad1)
        variance = param.params[self.name]["variance"]
        return variance * K * (r**self.order)

    def K2(
        self,
        param: Param,
        X: Float[Array, "N D"],
        X2: Optional[Float[Array, "M D"]] = None,
    ) -> Float[Array, "N M"]:
        """
        Evaluates the covariances element-wise between the all pairs in `X` and `X2`.

        NOTE: If `X2` is not specified, the method evaluates the covariances between pairs of `X`.

        Args:
            param: The `Param` initialised with the kernel.
            X: Input Array.
            X2: Input Array.

        Returns:
            The covariance between the specified intpus.
        """
        X2 = X if X2 is None else X2

        def _k_func(x1, x2):
            x_sphere, rad1 = self.to_sphere(param, x1)
            x_sphere2, rad2 = self.to_sphere(param, x2)
            k = jnp.dot(x_sphere, x_sphere2)
            k = self.shape_function(param, k)
            r = rad1 * rad2
            variance = param.params[self.name]["variance"]
            return variance * k * (r**self.order)

        return jax.vmap(lambda x: jax.vmap(lambda y: _k_func(x, y))(X2))(X)

    def K_diag(self, param: Param, X: Float[Array, "N D"]) -> Float[Array, " N"]:
        """
        Evaluates the diagonal of the kernel, i.e., the variances of the input.

        NOTE: We don't need to evaluate the shape function, as the cosine for the diagonal elements
        is 1 and the shape function is normalised so that `κ(1) = 1`.

        Args:
            param: The `Param` initialised with the kernel.
            X: Input Array.

        Returns:
            The diagonal variances of specified intpus.
        """
        _, rad = self.to_sphere(param, X)
        variance = param.params[self.name]["variance"]
        return variance * rad ** (2 * self.order)

    # def __add__(self, other: "Spherical") -> "Spherical":
    #     return Sum(kernels=[self, other])  # type: ignore

    # def __mul__(self, other: "Spherical") -> "Spherical":
    #     return Product(kernels=[self, other])  # type: ignore


class ArcCosine(Spherical):
    """
    The ArcCosine kernel.

    Attributes:
        depth: the depth of the equivalent neural network, which corresponds to nested computations
            of the `shape_function`. Defautls to 1.
        name: The name of the kernel. Defautls to `"ArcCosine"`.

    Raises:
        ValueError: if specified order is not in `{0, 1, 2}`.

    Returns:
        The kernel.
    """

    depth: int = field(default=1, pytree_node=False)
    name: str = field(default="ArcCosine", pytree_node=False)

    def __post_init__(self):
        """
        Check after the creation of the dataclass if we have provided a valid `order`.

        Raises:
            ValueError: if specified order is not in `{0, 1, 2}`.
        """
        if self.order not in {0, 1, 2}:
            raise ValueError("Requested order is not implemented.")

    def shape_function(self, param: Param, x: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        The shape function that determines the equivalent kernel to a deep network architecture.

        Args:
            param: The `Param` initialised with the kernel.
            x: Input Array.
        """
        x = self._squash(x)

        # step function for 1 layer
        def step(carry: Array, dummy: Optional[Array] = None) -> Tuple[Array, Array]:
            y = self.kappa(carry, self.order)
            return y, y

        # run scan for going deep.
        _, y = jax.lax.scan(step, x, xs=None, length=self.depth)
        return y[-1]


class NTK(Spherical):
    """
    The neural tangent kernel.

    Attributes:
        depth: the depth of the equivalent neural network, which corresponds to nested computations
            of the `shape_function`. Defautls to 1.
        name: The name of the kernel. Defautls to `"NTK"`.

    Returns:
        _description_
    """

    order: int = field(init=False, pytree_node=False)
    depth: int = field(default=1, pytree_node=False)
    name: str = field(default="NTK", pytree_node=False)

    def __post_init__(self):
        """Hard-wire the `order` to be 1."""
        object.__setattr__(self, "order", 1)

    def shape_function(self, param: Param, x: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        The shape function that determines the equivalent kernel to a deep network architecture.

        Args:
            param: The `Param` initialised with the kernel.
            x: Input Array.
        """
        x = self._squash(x)

        # step function for 1 layer
        def step(
            carry: Tuple[Array, Array], dummy: Optional[Array] = None
        ) -> Tuple[Tuple[Array, Array], Array]:
            x, y = carry
            x, y = self.kappa(x, 1), y * self.kappa(x, 0) + self.kappa(x, 1)
            carry = x, y
            return carry, y

        # run scan for going deep.
        _, y = jax.lax.scan(step, (x, x), xs=None, length=self.depth)
        return y[-1] / (self.depth + 1)


class PolynomialDecay(Spherical):
    """
    The polynomial decay kernel which parameterises the decay rate of the eigenvalues.

    The functional form is:

        K(u) = Σₙ n⁻ᵝ Cᵅₙ(u) (n + α) / α,

        where n is the ordinal frequency, α is a constant on the sphere and β the decay rate.

    Attributes:
        truncation_level: the order at which we truncate the frequencies.
        name: The name for the object. Defautls to `"PolynomialDecay"`.
    """

    truncation_level: int = field(default=10, pytree_node=False)
    order: int = field(init=False, pytree_node=False)
    name: str = field(default="PolynomialDecay", pytree_node=False)

    def __post_init__(self):
        """Hard-wire the `order` to be 1."""
        object.__setattr__(self, "order", 1)

    def _compute_eigvals(
        self,
        sphere_dim: int,
        *,
        max_num_eigvals: Optional[int] = None,
        gegenbauer_lookup_table: Optional[GegenbauerLookupTable] = None,
    ):
        """
        Compute the eigenvalues of the PolynomialDecay kernel.

        Without an initialised parameter we can only compute the constant part of the eigenvalues
            α / (n + α) / Cᵅₙ(1)

        Args:
            sphere_dim: The dimensionality of the sphere.
            max_num_eigvals: The maximum number of eigenvalues to precompute.
                This is ignored as the kernel uses the `self.truncation_level` instead.
            gegenbauer_lookup_table: A precomputed lookup table to evaluate the Gegenbauer
                polynomial. Defaults to None.

        Raises:
            ValueError: If `gegenbauer_lookup_table` is specified when we don't use a
                `PolynomialDecay` kernel.

        Returns:
            Array with the eigenvalues of the kernel.
        """
        if not gegenbauer_lookup_table:
            raise ValueError("Lookup table should be provided with `PolyDecay` kernel.")

        dim = sphere_dim + 1
        alpha = (dim - 2.0) / 2.0
        geg = lambda n: gegenbauer_lookup_table(n, alpha, jnp.array(1.0, dtype=jnp.float64))
        C_1 = jax.vmap(geg)(jnp.arange(self.truncation_level, dtype=jnp.int32))
        n = jnp.arange(self.truncation_level, dtype=jnp.float64)
        return alpha / (n + alpha) / C_1

    def eigenvalues(self, param: Param, levels: Int[Array, " N"]) -> Float[Array, " N"]:
        """
        Get the eigenvalues of the PolynomialDecay kernel for the specified levels.

        For the nth frequency we have::
            λₙ = (n + 1)⁻ᵝ α / (n + α) / Cᵅₙ(1) / Σₘ(m + 1)⁻ᵝ

        NOTE: we clip the provided levels to the `self.truncation_level`.

        Args:
            param: A `Param` initialised with the kernel.
            levels: An array specifying up to which order de we need eigenvalues.

        Returns:
            The eigenvalues.
        """
        n = jnp.arange(self.truncation_level, dtype=jnp.float64)
        beta = param.params[self.name]["beta"]
        sphere_dim = param.constants["sphere"]["sphere_dim"]
        geg = param.constants["sphere"]["gegenbauer_lookup_table"]
        decay = (1 + n) ** (-beta)
        const_factor = self._compute_eigvals(sphere_dim, gegenbauer_lookup_table=geg)
        eigval = decay * const_factor / jnp.sum(decay)
        levels = jnp.clip(levels, 0, self.truncation_level - 1)
        return eigval[levels]

    def shape_function(self, param: Param, x: Float[Array, "N D"]) -> Float[Array, "N D"]:
        """
        The shape function that determines the equivalent kernel to a deep network architecture.

        Args:
            param: The `Param` initialised with the kernel.
            x: Input Array.

        Raises:
            NotImplementedError: All derived classes need to implement this method.
        """
        x = self._squash(x)

        alpha = param.constants["sphere"]["alpha"]
        gegenbauer_lookup = param.constants["sphere"]["gegenbauer_lookup_table"]
        levels = jnp.arange(self.truncation_level, dtype=jnp.int32)

        const_factor = self.eigenvalues(param, levels)
        eigvals = const_factor * (levels.astype(jnp.float64) + alpha) / alpha
        C_n_x = jax.vmap(lambda n: gegenbauer_lookup(n, alpha, x))(levels)
        return jnp.sum(jax.vmap(lambda n, e: n * e)(C_n_x, eigvals), 0)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from gpfy.likelihoods import Bernoulli, Gaussian

key = jr.PRNGKey(42)


def test_init_create_correct_variables():
    lik = Gaussian()
    param = lik.init()

    assert "variance" in param.params["likelihood"][lik.name]


@pytest.mark.parametrize("likelihood_class", [Gaussian, Bernoulli])
def test_variational_expectations(likelihood_class):
    """
    Here we make sure that the variational_expectations gives the same result
    as log_prob if the latent function has no uncertainty.
    """
    lik = likelihood_class()
    param = lik.init()
    y = jr.normal(key, (10, 5))
    f = jr.normal(key, (10, 5))
    fvar = jnp.zeros_like(f)

    r1 = lik.log_prob(param, f, y)
    r2 = lik.variational_expectations(param, f, fvar)(y)
    np.testing.assert_allclose(r1, r2)



# Copyright 2023 Stefanos Eleftheriadis, James Hensman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
import optax
from jaxtyping import Array, Float

from gpfy.likelihoods import Likelihood
from gpfy.param import Param
from gpfy.scan import vscan
from gpfy.spherical import Spherical
from gpfy.spherical_harmonics import SphericalHarmonics
from gpfy.training import TrainState
from gpfy.typing import PRNG, TrainStepFn
from gpfy.utils import dataclass, field
from gpfy.variational import VariationalDistribution


@dataclass
class GP:
    """
    A Gaussian process model parameterised by a mean and a covariance function.

    Currently, mean funciton is not implemented so we assume all models to have zero mean.

    Atrirbutes:
        kernel: A spherical kernel acting as the covariance function of the GP.
        conditional_fn: A tuple of four functions that define how to do conditioning for the GP via:
            1. The projection function that computes the projection operation Kₘₘ⁻¹Kₘₙ.
            2. The conditional mean that computes the predictive mean given the projection.
            3. The conditional variance that computes the diagonal variances given the projection.
            4. The conditional covariance that computes the covariance given the porjection.

            It defaults to an empty tuple, which implies predicting from the prior GP.
        name: The name for the object. Defautls to the class name.

    Returns:
        A GP model.
    """

    kernel: Spherical = field(pytree_node=False)
    conditional_fn: Tuple[Callable, Callable, Callable, Callable] = field(
        default_factory=tuple, pytree_node=False
    )
    name: str = field(default="GP", pytree_node=False)

    def __init_subclass__(cls):
        dataclass(cls)  # pytype: disable=wrong-arg-types

    def init(
        self,
        key: PRNG,
        input_dim: int,
        num_independent_processes: int = 1,
        projection_dim: Optional[int] = None,
        likelihood: Optional[Likelihood] = None,
        sh_features: Optional[SphericalHarmonics] = None,
        variational_dist: Optional[VariationalDistribution] = None,
    ) -> Param:
        """
        Initialise the parameters of a sparse sphercial GP.

        Args:
            key: A random key for initialising the kernel and the spherical harmonics.
            input_dim: The input dimension. We then append an extra dimension for bias.
            num_independent_processes: The number of independent processes we need to learn.
                Used for independently modelling multiple outputs. Defaults to 1.
            projection_dim: If specified it denotes a projection of the `input_dim` to
                `projection_dim`. Defaults to None.
            likelihood: The likelihood to model the data. Defaults to None.
            sh_features: The spherical harmonics features for sparse inference. Defaults to None.
            variational_dist: The variational distribution to approximate the posterior process.
                Defaults to None.

        Returns:
            The `Param` object with all variables.
        """
        key, subkey = jax.random.split(key)
        param = self.kernel.init(subkey, input_dim, projection_dim)
        param = likelihood.init(param) if likelihood else param
        if variational_dist and sh_features:
            param = sh_features.init(key, projection_dim or input_dim, param)
            param = variational_dist.init(
                sh_features.num_inducing(param), num_independent_processes, param
            )

        return param

    def conditional(
        self,
        sh: SphericalHarmonics,
        q: VariationalDistribution,
    ) -> "GP":
        """
        Condition the process on the specified pseudo observations, i.e., the inducing features.

        Args:
            sh: The spherical harmonic features.
            q: The variational distribution.

        Returns:
            A new GP model with populated `self.conditional_fn` to make predictions from.
        """
        proj_fun, cond_cov, cond_var = sh.conditional_fun(self.kernel)

        # cond_var = lambda param, x, proj: self.kernel.K_diag(param, x) - jnp.sum(
        #     jnp.square(proj), -2
        # )
        # cond_cov = lambda param, x, proj: self.kernel.K(param, x) - jnp.matmul(
        #     proj.swapaxes(-1, -2), proj
        # )

        # conditional mean function
        cond_mean_fun = lambda param, proj: q.project_mean(param, proj)

        # conditional covariance function
        cond_cov_fun = lambda param, x, proj: cond_cov(param, x, proj) + q.project_variance(
            param, proj
        )

        # conditional diagonal variance function
        cond_var_fun = lambda param, x, proj: cond_var(param, x, proj)[
            ..., None
        ] + q.project_diag_variance(param, proj)
        conditional_fn = (
            (proj_fun),
            (cond_mean_fun),
            (cond_cov_fun),
            (cond_var_fun),
        )
        return GP(self.kernel, conditional_fn)

    def mu(self, param: Param) -> Callable[[Float[Array, "N Din"]], Float[Array, "N Dout"]]:
        """
        The mean function of the GP.

        If the `self.conditional_fn` tuple is defined, then it returns the specified predictive
        mean function. Otherwise it returns the prior mean function.

        Args:
            param: A `Param` initialised with the model.

        Returns:
            The conditional mean function or the prior mean.
        """
        if self.conditional_fn:
            proj_fun, cond_mean_fun, _, __ = self.conditional_fn
            return lambda x: cond_mean_fun(param, proj_fun(param, x))
        return lambda x: jnp.zeros((x.shape[0], 1), dtype=jnp.float64)

    def var(self, param: Param) -> Callable[[Float[Array, "N Din"]], Float[Array, "N Dout"]]:
        """
        The diagonal variance function of the GP.

        If the `self.conditional_fn` tuple is defined, then it returns the specified diagonal
        predictive variance function. Otherwise it returns the prior variance.

        Args:
            param: A `Param` initialised with the model.

        Returns:
            The conditional variance function or the prior variance.
        """
        if self.conditional_fn:
            proj_fun, _, __, cond_var_fun = self.conditional_fn
            return lambda x: cond_var_fun(param, x, proj_fun(param, x))
        return lambda x: self.kernel.K_diag(param, x)

    def cov(self, param: Param) -> Callable[[Float[Array, "N D"]], Float[Array, "N N"]]:
        """
        The covariance function of the GP.

        If the `self.conditional_fn` tuple is defined, then it returns the specified predictive
        covariance function. Otherwise it returns the prior covariance.

        Args:
            param: A `Param` initialised with the model.

        Returns:
            The conditional covariance function or the prior covariance.
        """
        if self.conditional_fn:
            proj_fun, _, cond_cov_fun, __ = self.conditional_fn
            return lambda x: cond_cov_fun(param, x, proj_fun(param, x))
        return lambda x: self.kernel.K(param, x)

    def predict_diag(
        self, param: Param, X: Float[Array, "N Din"]
    ) -> Tuple[Float[Array, "N Dout"], Float[Array, "N Dout"]]:
        """
        Predict the mean and the diagonal variance of the GP, given some input `X`.

        If the `self.conditional_fn` tuple is defined, then it predicts from the equivalent
        conditional distribution. Otherwise it predicts from the prior process.

        NOTE: We explicitly duplicate code here instead of directly calling `self.mu` and aself.var`
        to avoid unecessary duplicate operations.

        Args:
            param: A `Param` initialised with the model.
            X: Input array.

        Returns:
            The predictive mean and diagonal variance of the (conditional) process.
        """
        if self.conditional_fn:
            proj_fun, cond_mean_fun, _, cond_var_fun = self.conditional_fn

            proj = proj_fun(param, X)
            mu = cond_mean_fun(param, proj)
            var = cond_var_fun(param, X, proj)
        else:
            mu = jnp.zeros((X.shape[0], 1), dtype=jnp.float64)
            var = self.kernel.K_diag(param, X)[..., None]
        return mu, var

    def predict(
        self, param: Param, X: Float[Array, "N Din"]
    ) -> Tuple[Float[Array, "N Dout"], Float[Array, "N N"]]:
        """
        Predict the mean and the covariance of the GP, given some input `X`.

        If the `self.conditional_fn` tuple is defined, then it predicts from the equivalent
        conditional distribution. Otherwise it predicts from the prior process.

        NOTE: We explicitly duplicate code here instead of directly calling `self.mu` and aself.var`
        to avoid unecessary duplicate operations.

        Args:
            param: A `Param` initialised with the model.
            X: Input array.

        Returns:
            The predictive mean and covariance of the (conditional) process.
        """
        if self.conditional_fn:
            proj_fun, cond_mean_fun, cond_cov_fun, _ = self.conditional_fn

            proj = proj_fun(param, X)
            mu = cond_mean_fun(param, proj)
            var = cond_cov_fun(param, X, proj)
        else:
            mu = jnp.zeros((X.shape[0], 1), dtype=jnp.float64)
            var = self.kernel.K(param, X)
        return mu, var

    def fit(
        self,
        param: Param,
        train_step: TrainStepFn,
        optimizer: optax.GradientTransformation,
        num_iters: int,
        progress_bar: bool = True,
    ):
        """
        Optimise the parameters to fit the GP model on the training data.

        Args:
            param: A `Param` initialised with the model.
            train_step: A training step function that is called inside a loop during optimisation.
            optimizer: The optimizer to use.
            num_iters: The number of iterations.
            progress_bar: Flag to indicate if we want to print a progress bar during optimisation.
                Defaults to True.

        Returns:
            A tuple of 3 objects containing the optimised `Param`, the final `TrainingState` and
            an `Array` containing the value of the negative `elbo` during the optimisation steps.
        """
        # Get the unconstrained params
        param_free = param.unconstrained()

        # Get the non-trainable params and set the optimizer to zero_grad
        frozen = jax.tree_map(lambda x: not (x), param_free._trainables)
        tx = optax.chain(optimizer, optax.masked(optax.set_to_zero(), frozen))

        # initialise a train state with the free parameters and an `apply_fn` that replaces the
        # VariableDict in the free `Param`.
        state = TrainState.create(
            apply_fn=lambda p: param_free.replace(params=p), params=param_free.params, tx=tx
        )

        scan: Callable = vscan if progress_bar else jax.lax.scan  # type: ignore
        state, loss_val = scan(train_step, state, None, num_iters)

        param_new = state.apply_fn(state.params).constrained()

        return param_new, state, loss_val
